import numpy as np
from numpy.testing import assert_allclose

import pytest
from tomophantom.artefacts import artefacts_mix

eps = 1e-05


@pytest.mark.parametrize("noise_type", ["Poisson", "Gaussian"])
def test_2d_sino_noise(noise_type, sino_model11_64):
    sino_an, angles_num, P = sino_model11_64
    ideal_sino_sum = np.sum(sino_an)
    if noise_type == "Poisson":
        _noise_ = {
            "noise_type": noise_type,
            "noise_amplitude": 10000,
            "noise_seed": 1,
        }
    else:
        _noise_ = {
            "noise_type": noise_type,
            "noise_amplitude": 0.1,
            "noise_seed": 1,
        }
    noisy_sino = artefacts_mix(sino_an, **_noise_)
    noisy_sino_sum = np.sum(noisy_sino)
    assert noisy_sino_sum != ideal_sino_sum
    assert 0.0 <= np.max(noisy_sino) <= 500
    assert noisy_sino.dtype == np.float32
    assert noisy_sino.shape == (angles_num, P)


def test_2d_sino_zingers(sino_model11_64):
    sino_an, angles_num, P = sino_model11_64
    ideal_sino_sum = np.sum(sino_an)

    _zingers_ = {"zingers_percentage": 2, "zingers_modulus": 10}

    noisy_sino = artefacts_mix(sino_an, **_zingers_)
    noisy_sino_sum = np.sum(noisy_sino)
    assert noisy_sino_sum != ideal_sino_sum
    assert 0.0 <= np.max(noisy_sino) <= 500
    assert noisy_sino.dtype == np.float32
    assert noisy_sino.shape == (angles_num, P)


def test_2d_sino_stripes(sino_model11_64):
    sino_an, angles_num, P = sino_model11_64
    ideal_sino_sum = np.mean(sino_an)

    _stripes_ = {
        "stripes_percentage": 0.8,
        "stripes_maxthickness": 2.0,
        "stripes_intensity": 0.25,
        "stripes_type": "full",
        "stripes_variability": 0.002,
    }

    noisy_sino = artefacts_mix(sino_an, **_stripes_)
    noisy_sino_sum = np.sum(np.abs(noisy_sino - ideal_sino_sum))
    assert noisy_sino_sum > 0.0
    assert 0.0 <= np.max(noisy_sino) <= 500
    assert noisy_sino.dtype == np.float32
    assert noisy_sino.shape == (angles_num, P)


def test_2d_sino_shifts_pixel(sino_model11_64):
    sino_an, angles_num, P = sino_model11_64

    _sinoshifts_ = {"datashifts_maxamplitude_pixel": 10}

    [noisy_sino_misalign, shifts] = artefacts_mix(sino_an, **_sinoshifts_)
    noisy_sino_sum = np.sum(np.abs(noisy_sino_misalign - sino_an))
    assert noisy_sino_sum > 0.0
    assert 0.0 <= np.max(noisy_sino_misalign) <= 500
    assert noisy_sino_misalign.dtype == np.float32
    assert noisy_sino_misalign.shape == (angles_num, P)


def test_2d_sino_shifts_subpixel(sino_model11_64):
    sino_an, angles_num, P = sino_model11_64

    _sinoshifts_ = {"datashifts_maxamplitude_subpixel": 0.7}

    [noisy_sino_misalign, shifts] = artefacts_mix(sino_an, **_sinoshifts_)
    noisy_sino_sum = np.sum(np.abs(noisy_sino_misalign - sino_an))
    assert noisy_sino_sum > 0.0
    assert 0.0 <= np.max(noisy_sino_misalign) <= 500
    assert noisy_sino_misalign.dtype == np.float32
    assert noisy_sino_misalign.shape == (angles_num, P)


def test_2d_sino_pve(sino_model11_64):
    sino_an, angles_num, P = sino_model11_64

    _pve_ = {"pve_strength": 3}

    noisy_sino = artefacts_mix(sino_an, **_pve_)
    noisy_sino_sum = np.sum(np.abs(noisy_sino - sino_an))
    assert noisy_sino_sum > 0.0
    assert 0.0 <= np.max(noisy_sino) <= 500
    assert noisy_sino.dtype == np.float32
    assert noisy_sino.shape == (angles_num, P)


def test_2d_sino_frenel(sino_model11_64):
    sino_an, angles_num, P = sino_model11_64

    _fresnel_ = {
        "fresnel_dist_observation": 50,
        "fresnel_scale_factor": 10,
        "fresnel_wavelenght": 0.001,
    }

    noisy_sino = artefacts_mix(sino_an, **_fresnel_)
    noisy_sino_sum = np.sum(np.abs(noisy_sino - sino_an))
    assert noisy_sino_sum > 0.0
    assert 0.0 <= np.max(noisy_sino) <= 500
    assert noisy_sino.dtype == np.float32
    assert noisy_sino.shape == (angles_num, P)


def test_2d_sino_combi(sino_model11_64):
    sino_an, angles_num, P = sino_model11_64

    _noise_ = {
        "noise_type": "Poisson",
        "noise_amplitude": 10000,
        "noise_seed": 1,
    }
    _zingers_ = {"zingers_percentage": 2, "zingers_modulus": 10}

    _stripes_ = {
        "stripes_percentage": 0.8,
        "stripes_maxthickness": 2.0,
        "stripes_intensity": 0.25,
        "stripes_type": "full",
        "stripes_variability": 0.002,
    }

    _sinoshifts_ = {"datashifts_maxamplitude_subpixel": 0.7}

    _pve_ = {"pve_strength": 3}

    _fresnel_ = {
        "fresnel_dist_observation": 50,
        "fresnel_scale_factor": 10,
        "fresnel_wavelenght": 0.001,
    }

    [noisy_sino, shifts] = artefacts_mix(
        sino_an,
        **_noise_,
        **_zingers_,
        **_stripes_,
        **_sinoshifts_,
        **_pve_,
        **_fresnel_,
    )

    noisy_sino_sum = np.sum(np.abs(noisy_sino - sino_an))
    assert noisy_sino_sum > 0.0
    assert 0.0 <= np.max(noisy_sino) <= 500
    assert noisy_sino.dtype == np.float32
    assert noisy_sino.shape == (angles_num, P)



import numpy as np
from numpy.testing import assert_allclose

import pytest
import os
import tomophantom
from tomophantom.TomoP3D import (
    _check_params3d,
    Model,
    ModelSub,
    ModelTemporal,
    ModelTemporalSub,
)

eps = 1e-05


@pytest.mark.parametrize("model", list(np.arange(1, 19, 1)))
def test_3d_model(model):
    N_size = 64
    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library3D = os.path.join(path, "phantomlib", "Phantom3DLibrary.dat")

    # check parameters
    params = _check_params3d(model, path_library3D)
    phantom_3D = Model(model, N_size, path_library3D)

    assert np.sum(params) == 12
    assert params.dtype == np.int32
    assert 0.001 <= np.max(phantom_3D) <= 100
    assert phantom_3D.dtype == np.float32
    assert phantom_3D.shape == (N_size, N_size, N_size)


def test_3d_model_sm():
    phantom_size = (64, 32, 64)
    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library3D = os.path.join(path, "phantomlib", "Phantom3DLibrary.dat")

    # check parameters
    params = _check_params3d(11, path_library3D)
    phantom_3D = Model(11, phantom_size, path_library3D)

    assert np.sum(params) == 12
    assert params.dtype == np.int32
    assert 0.001 <= np.max(phantom_3D) <= 100
    assert phantom_3D.dtype == np.float32
    assert phantom_3D.shape == phantom_size


def test_3d_model_sub():
    phantom_size = (64, 64, 64)
    subset_ind = (15, 30)
    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library3D = os.path.join(path, "phantomlib", "Phantom3DLibrary.dat")

    # check parameters
    params = _check_params3d(11, path_library3D)
    phantom_3D = ModelSub(11, phantom_size, subset_ind, path_library3D)

    assert np.sum(params) == 12
    assert params.dtype == np.int32
    assert 0.001 <= np.max(phantom_3D) <= 100
    assert phantom_3D.dtype == np.float32
    assert phantom_3D.shape == (15, 64, 64)


@pytest.mark.parametrize("model", list(np.arange(100, 103, 1)))
def test_3d_model_temporal(model):
    N_size = 64
    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library3D = os.path.join(path, "phantomlib", "Phantom3DLibrary.dat")

    # check parameters
    params = _check_params3d(model, path_library3D)
    phantom_4D = ModelTemporal(model, N_size, path_library3D)

    assert len(params) == 12
    assert params.dtype == np.int32
    assert 0.001 <= np.max(phantom_4D) <= 100
    assert phantom_4D.dtype == np.float32
    assert phantom_4D.shape == (params[3], N_size, N_size, N_size)


def test_3d_temp_model_sub():
    phantom_size = (64, 64, 64)
    subset_ind = (15, 30)
    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library3D = os.path.join(path, "phantomlib", "Phantom3DLibrary.dat")

    # check parameters
    params = _check_params3d(100, path_library3D)
    phantom_3D = ModelTemporalSub(100, phantom_size, subset_ind, path_library3D)

    assert len(params) == 12
    assert params.dtype == np.int32
    assert 0.001 <= np.max(phantom_3D) <= 100
    assert phantom_3D.dtype == np.float32
    assert phantom_3D.shape == (params[3], 15, 64, 64)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module for internal utility functions.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import ctypes
import logging
import multiprocessing as mp

import numpy as np

logger = logging.getLogger(__name__)

__all__ = [
    "as_ndarray",
    "as_dtype",
    "as_float32",
    "as_int32",
    "as_uint8",
    "as_uint16",
    "as_c_float_p",
    "as_c_bool_p",
    "as_c_uint8_p",
    "as_c_uint16_p",
    "as_c_int",
    "as_c_int_p",
    "as_c_float",
    "as_c_char_p",
    "as_c_void_p",
    "as_c_size_t",
]


def as_ndarray(arr, dtype=None, copy=False):
    if not isinstance(arr, np.ndarray):
        arr = np.array(arr, dtype=dtype, copy=copy)
    return arr


def as_dtype(arr, dtype, copy=False):
    if not arr.dtype == dtype:
        arr = np.array(arr, dtype=dtype, copy=copy)
    return arr


def as_float32(arr):
    arr = as_ndarray(arr, np.float32)
    return as_dtype(arr, np.float32)


def as_int32(arr):
    arr = as_ndarray(arr, np.int32)
    return as_dtype(arr, np.int32)


def as_uint16(arr):
    arr = as_ndarray(arr, np.uint16)
    return as_dtype(arr, np.uint16)


def as_uint8(arr):
    arr = as_ndarray(arr, np.uint8)
    return as_dtype(arr, np.uint8)


def as_c_float_p(arr):
    c_float_p = ctypes.POINTER(ctypes.c_float)
    return arr.ctypes.data_as(c_float_p)


def as_c_bool_p(arr):
    c_bool_p = ctypes.POINTER(ctypes.c_bool)
    return arr.ctypes.data_as(c_bool_p)


def as_c_uint8_p(arr):
    c_uint8_p = ctypes.POINTER(ctypes.c_uint8)
    return arr.ctypes.data_as(c_uint8_p)


def as_c_uint16_p(arr):
    c_uint16_p = ctypes.POINTER(ctypes.c_uint16)
    return arr.ctypes.data_as(c_uint16_p)


def as_c_int(arr):
    return ctypes.c_int(arr)


def as_c_long(arr):
    return ctypes.c_long(arr)


def as_c_int_p(arr):
    arr = arr.astype(np.intc, copy=False)
    c_int_p = ctypes.POINTER(ctypes.c_int)
    return arr.ctypes.data_as(c_int_p)


def as_c_float(arr):
    return ctypes.c_float(arr)


def as_c_char_p(arr):
    return ctypes.c_char_p(arr.encode())


def as_c_void_p():
    return ctypes.POINTER(ctypes.c_void_p)


def as_c_size_t(arr):
    return ctypes.c_size_t(arr)


def to_numpy_array(obj, dtype, shape):
    return np.frombuffer(obj, dtype=dtype).reshape(shape)


def is_contiguous(arr):
    return arr.flags.c_contiguous



import numpy as np
from numpy.testing import assert_allclose

import pytest
import os
import tomophantom
from tomophantom.TomoP2D import (
    _check_params2d,
    Model,
    ModelSino,
    SinoNum,
    ModelSinoTemporal,
)

eps = 1e-05


@pytest.mark.parametrize("model", list(np.arange(1, 16, 1)))
def test_2d_models_sino(model):
    N_size = 64  # set the desired dimension of the phantom
    angles_num = int(0.5 * np.pi * N_size)
    # angles number
    angles = np.linspace(0.0, 179.9, angles_num, dtype="float32")
    P = int(np.sqrt(2) * N_size)  # detectors

    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library2D = os.path.join(path, "phantomlib", "Phantom2DLibrary.dat")

    # Generate a N_size x N_size phantom (2D)
    sino_an = ModelSino(model, N_size, P, angles, path_library2D)
    assert 0.0 <= np.max(sino_an) <= 1000
    assert sino_an.dtype == np.float32
    assert sino_an.shape == (angles_num, P)


def test_num_sino_2d():
    N_size = 64  # set the desired dimension of the phantom
    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library2D = os.path.join(path, "phantomlib", "Phantom2DLibrary.dat")

    angles_num = int(0.5 * np.pi * N_size)
    # angles number
    angles = np.linspace(0.0, 179.9, angles_num, dtype="float32")
    P = int(np.sqrt(2) * N_size)  # detectors

    phantom_2D = Model(1, N_size, path_library2D)
    sino_num = SinoNum(phantom_2D, P, angles)

    assert sino_num.dtype == np.float32
    assert sino_num.shape == (100, 90)


@pytest.mark.parametrize("model", list(np.arange(100, 103, 1)))
def test_2d_temporal_sinos(model):
    N_size = 64  # set the desired dimension of the phantom
    angles_num = int(0.5 * np.pi * N_size)
    # angles number
    angles = np.linspace(0.0, 179.9, angles_num, dtype="float32")
    P = int(np.sqrt(2) * N_size)  # detectors

    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library2D = os.path.join(path, "phantomlib", "Phantom2DLibrary.dat")

    params = _check_params2d(model, path_library2D)
    # Generate a time_frame x N_size x N_size phantom (3D)
    sino2d_t = ModelSinoTemporal(model, N_size, P, angles, path_library2D)
    assert 0.001 <= np.max(sino2d_t) <= 1000
    assert sino2d_t.dtype == np.float32
    assert sino2d_t.shape == (params[3], P, angles_num)



import numpy as np
from numpy.testing import assert_allclose

import pytest
import os
import tomophantom
from tomophantom.TomoP3D import (
    _check_params3d,
    ModelSino,
    ModelSinoSub,
    ModelSinoTemporal,
    ModelSinoTemporalSub,
)

eps = 1e-05


@pytest.mark.parametrize("model", list(np.arange(1, 19, 1)))
def test_3d_models_sino(model):
    N_size = 64  # set the desired dimension of the phantom
    # Projection geometry related parameters:
    Horiz_det = int(np.sqrt(2) * N_size)  # detector column count (horizontal)
    Vert_det = N_size  # detector row count (vertical) (no reason for it to be > N)
    angles_num = int(0.5 * np.pi * N_size)
    # angles number
    angles = np.linspace(0.0, 179.9, angles_num, dtype="float32")  # in degrees

    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library3D = os.path.join(path, "phantomlib", "Phantom3DLibrary.dat")

    # Generatea 3d sino
    sino3d = ModelSino(model, N_size, Horiz_det, Vert_det, angles, path_library3D)
    assert 0.0 <= np.max(sino3d) <= 1000
    assert sino3d.dtype == np.float32
    assert sino3d.shape == (Vert_det, angles_num, Horiz_det)


def test_3d_models_sino_sub():
    N_size = 64  # set the desired dimension of the phantom
    # Projection geometry related parameters:
    Horiz_det = int(np.sqrt(2) * N_size)  # detector column count (horizontal)
    Vert_det = N_size  # detector row count (vertical) (no reason for it to be > N)
    angles_num = int(0.5 * np.pi * N_size)
    # angles number
    angles = np.linspace(0.0, 179.9, angles_num, dtype="float32")  # in degrees

    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library3D = os.path.join(path, "phantomlib", "Phantom3DLibrary.dat")
    subset_tuple = (15, 30)

    # Generate a 3d sino
    sino3d = ModelSinoSub(
        11, N_size, Horiz_det, Vert_det, subset_tuple, angles, path_library3D
    )
    assert 0.0 <= np.max(sino3d) <= 1000
    assert sino3d.dtype == np.float32
    assert sino3d.shape == (15, angles_num, Horiz_det)


@pytest.mark.parametrize("model", list(np.arange(100, 103, 1)))
def test_3d_models_sino_temporal(model):
    N_size = 64  # set the desired dimension of the phantom
    # Projection geometry related parameters:
    Horiz_det = int(np.sqrt(2) * N_size)  # detector column count (horizontal)
    Vert_det = N_size  # detector row count (vertical) (no reason for it to be > N)
    angles_num = int(0.5 * np.pi * N_size)
    # angles number
    angles = np.linspace(0.0, 179.9, angles_num, dtype="float32")  # in degrees

    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library3D = os.path.join(path, "phantomlib", "Phantom3DLibrary.dat")

    params = _check_params3d(model, path_library3D)

    # Generatea 4d sino
    sino4d = ModelSinoTemporal(
        model, N_size, Horiz_det, Vert_det, angles, path_library3D
    )
    assert 0.0 <= np.max(sino4d) <= 1000
    assert sino4d.dtype == np.float32
    assert sino4d.shape == (params[3], Vert_det, angles_num, Horiz_det)


def test_3d_models_sino_temporal_sub():
    N_size = 64  # set the desired dimension of the phantom
    model = 100
    # Projection geometry related parameters:
    Horiz_det = int(np.sqrt(2) * N_size)  # detector column count (horizontal)
    Vert_det = N_size  # detector row count (vertical) (no reason for it to be > N)
    angles_num = int(0.5 * np.pi * N_size)
    # angles number
    angles = np.linspace(0.0, 179.9, angles_num, dtype="float32")  # in degrees

    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library3D = os.path.join(path, "phantomlib", "Phantom3DLibrary.dat")
    params = _check_params3d(model, path_library3D)
    subset_tuple = (15, 30)

    # Generate a 3d sino
    sino4d = ModelSinoTemporalSub(
        model, N_size, Horiz_det, Vert_det, subset_tuple, angles, path_library3D
    )
    assert 0.0 <= np.max(sino4d) <= 1000
    assert sino4d.dtype == np.float32
    assert sino4d.shape == (params[3], 15, angles_num, Horiz_det)



import numpy as np
from numpy.testing import assert_allclose

from tomophantom.generator import foam2D, foam3D

eps = 1e-05


def test_2d_generator():
    N_size = 64  # set the desired dimension of the phantom
    tot_objects = 20  # the total number of objects to generate

    # define ranges for parameters
    x0min = -0.9
    x0max = 0.9
    y0min = -0.9
    y0max = 0.9
    c0min = 0.01
    c0max = 1.0
    ab_min = 0.01
    ab_max = 0.25

    # 2D example
    (Objfoam2D, myObjects) = foam2D(
        x0min,
        x0max,
        y0min,
        y0max,
        c0min,
        c0max,
        ab_min,
        ab_max,
        N_size,
        tot_objects,
        object_type="mix",
    )

    assert 0.0 <= np.max(Objfoam2D) <= 10
    assert Objfoam2D.dtype == np.float32
    assert Objfoam2D.shape == (N_size, N_size)


def test_3d_generator():
    N_size = 64  # define the grid size
    tot_objects = 10  # the total number of objects to generate

    # define ranges for parameters
    x0min = -0.9
    x0max = 0.9
    y0min = -0.9
    y0max = 0.9
    z0min = -0.9
    z0max = 0.9
    c0min = 0.01
    c0max = 1.0
    ab_min = 0.01
    ab_max = 0.25

    (Objfoam3D, myObjects) = foam3D(
        x0min,
        x0max,
        y0min,
        y0max,
        z0min,
        z0max,
        c0min,
        c0max,
        ab_min,
        ab_max,
        N_size,
        tot_objects,
        object_type="mix",
    )

    assert 0.0 <= np.max(Objfoam3D) <= 10
    assert Objfoam3D.dtype == np.float32
    assert Objfoam3D.shape == (N_size, N_size, N_size)



"""Modules for generating 3D synthetic phantoms and their analytical projection data.
The dynamic phantoms (resulting in 4D arrays) can also be generated.

API Summary:

* :func:`Model` - generates a 3D phantom using a :ref:`ref_glossary_model` from :ref:`howto_3d_libs`.
* :func:`ModelSub` - generates a vertical subset (a cutoff) for a :ref:`ref_glossary_model` from :ref:`howto_3d_libs`.
* :func:`ModelTemporal` - generates a :ref:`ref_glossary_dynamic_model` (3D + t) from the from :ref:`howto_3d_libs`.
* :func:`ModelTemporalSub` - generates a vertical subset (cutoff) of a 4D :ref:`ref_glossary_dynamic_model` from the :ref:`howto_3d_libs`.
* :func:`ModelSino` - generates 3D projection data for a :ref:`ref_glossary_model` from the :ref:`howto_3d_libs`.
* :func:`ModelSinoSub` - generates a vertical subset (cutoff) for 3D projection data for a :ref:`ref_glossary_model` from the :ref:`howto_3d_libs`.
* :func:`ModelSinoTemporal` - generates (3D + t) 4D projection data for a temporal model from the :ref:`howto_3d_libs`.
* :func:`ModelSinoTemporalSub` - generates (3D + t) 4D projection data (subset) for a temporal model from the :ref:`howto_3d_libs`.
* :func:`Object` - generates one elementary 3D :ref:`ref_glossary_object` from the provided parameters, see :ref:`ref_object_api`.
* :func:`ObjectSino` - generates a 3D projection data for one :ref:`ref_glossary_object` from the provided parameters, see :ref:`ref_object_api`.
"""

import ctypes
import numpy as np
from numbers import Number
from enum import Enum
from typing import Union, Tuple
from pathlib import Path

import tomophantom.ctypes.external as external

__all__ = [
    "Model",
    "ModelSub",
    "ModelTemporal",
    "ModelTemporalSub",
    "ModelSino",
    "ModelSinoSub",
    "ModelSinoTemporal",
    "ModelSinoTemporalSub",
    "Object",
    "ObjectSino",
]


class Objects3D(Enum):
    """Enumeration with the available objects for 3D phantoms"""

    GAUSSIAN = "gaussian"
    PARABOLOID = "paraboloid"
    ELLIPSOID = "ellipsoid"
    CONE = "cone"
    CUBOID = "cuboid"
    ELLIPCYLINDER = "elliptical_cylinder"


def _check_params3d(model_no: int, models_library_path: Path) -> np.ndarray:
    """Check parameters before executing the generation script.

    Args:
        model_no (int): Model number from the Phantom3DLibrary.dat library file.
        models_library_path (Path): A path to the library file.

    Returns:
        list: a list of integers
    """
    params = np.ascontiguousarray(np.zeros([12], dtype=ctypes.c_int))
    external.c_checkParams3D(
        params,
        model_no,
        models_library_path,
    )
    return params


def Model(
    model_no: int,
    phantom_size: Union[int, Tuple[int, int, int]],
    models_library_path: Path,
) -> np.ndarray:
    """Generates a 3D phantom based on the model no. in the library file Phantom3DLibrary.dat.

    Args:
        model_no (int): Model number from the Phantom3DLibrary.dat library file.
        phantom_size (int, Tuple(int)): A scalar or a tuple with phantom dimensions.
        models_library_path (Path): A path to the library file.

    Returns:
        np.ndarray: The generated 3D phantom.
    """
    if type(phantom_size) == tuple:
        N1, N2, N3 = [int(i) for i in phantom_size]
    else:
        N1 = N2 = N3 = phantom_size

    # check the validity of model's parameters
    params = np.ascontiguousarray(np.zeros([12], dtype=ctypes.c_int))
    params = _check_params3d(model_no, models_library_path)
    __testParams3D(params)  # check parameters and terminate if incorrect

    if params[3] == 1:
        phantom = np.zeros([N1, N2, N3], dtype="float32", order="C")
    else:
        raise ValueError(
            "The selected model is temporal (3D+time), use 'ModelTemporal' function instead"
        )
    external.c_model3d(
        np.ascontiguousarray(phantom),
        model_no,
        N1,
        N2,
        N3,
        0,
        N1,
        models_library_path,
    )
    return phantom


def ModelSub(
    model_no: int,
    phantom_size: Union[int, Tuple[int, int, int]],
    sub_index: Tuple[int, int],
    models_library_path: Path,
) -> np.ndarray:
    """Generates a subset of 3D phantom (vertical cutoff) based on the model no. in
    the library file Phantom3DLibrary.dat.

    Args:
        model_no (int): Model number from the Phantom3DLibrary.dat library file.
        phantom_size (int, Tuple(int)): A scalar or a tuple with phantom dimensions.
        sub_index (Tuple[int, int]): a tuple containing 2 indeces (lower, upper) to select a vertical subset.
        models_library_path (Path): A path to the library file.

    Returns:
        np.ndarray: The generated 3D phantom.
    """
    if type(phantom_size) == tuple:
        N1, N2, N3 = [int(i) for i in phantom_size]
    else:
        N1 = N2 = N3 = phantom_size

    Z1, Z2 = [int(i) for i in sub_index]

    rangecheck = Z2 > Z1
    if not rangecheck:
        raise ValueError("Upper index must be larger than the lower one")
    rangecheck = Z1 >= 0 and Z1 < N3
    if not rangecheck:
        raise ValueError("Range of the lower index is incorrect")
    rangecheck = Z2 >= 0 and Z2 <= N3
    if not rangecheck:
        raise ValueError("Range of the higher index is incorrect")

    sub_size = Z2 - Z1  # the size of the vertical slab

    # check the validity of model's parameters
    params = np.ascontiguousarray(np.zeros([12], dtype=ctypes.c_int))
    params = _check_params3d(model_no, models_library_path)
    __testParams3D(params)  # check parameters and terminate if incorrect

    if params[3] == 1:
        phantom = np.zeros([sub_size, N2, N3], dtype="float32", order="C")
    else:
        raise ValueError(
            "The selected model is temporal (3D+time), use 'ModelTemporal' function instead"
        )
    external.c_model3d(
        np.ascontiguousarray(phantom),
        model_no,
        N1,
        N2,
        N3,
        Z1,
        Z2,
        models_library_path,
    )
    return phantom


def ModelTemporal(
    model_no: int,
    phantom_size: Union[int, Tuple[int, int, int]],
    models_library_path: Path,
) -> np.ndarray:
    """Generates 4D (3D+time) phantom based on the model no. in
    the library file Phantom3DLibrary.dat.

    Args:
        model_no (int): Model number from the Phantom3DLibrary.dat library file.
        phantom_size (int, Tuple(int)): A scalar or a tuple with phantom dimensions.
        models_library_path (Path): A path to the library file.

    Returns:
        np.ndarray: The generated 4D phantom.
    """
    if type(phantom_size) == tuple:
        N1, N2, N3 = [int(i) for i in phantom_size]
    else:
        N1 = N2 = N3 = phantom_size

    # check the validity of model's parameters
    params = np.ascontiguousarray(np.zeros([12], dtype=ctypes.c_int))
    params = _check_params3d(model_no, models_library_path)
    __testParams3D(params)  # check parameters and terminate if incorrect

    if params[3] == 1:
        raise ValueError(
            "The selected model is stationary (3D), use 'Model' function instead"
        )
    else:
        phantom = np.zeros([params[3], N1, N2, N3], dtype="float32", order="C")
    external.c_model3d(
        np.ascontiguousarray(phantom),
        model_no,
        N1,
        N2,
        N3,
        0,
        N1,
        models_library_path,
    )
    return phantom


def ModelTemporalSub(
    model_no: int,
    phantom_size: Union[int, Tuple[int, int, int]],
    sub_index: Tuple[int, int],
    models_library_path: Path,
) -> np.ndarray:
    """Generates a subset of 4D phantom (vertical cutoff) based on the model no. in
    the library file Phantom3DLibrary.dat.

    Args:
        model_no (int): Model number from the Phantom3DLibrary.dat library file.
        phantom_size (int, Tuple(int)): A scalar or a tuple with phantom dimensions.
        sub_index (Tuple[int, int]): a tuple containing 2 indeces (lower, upper) to select a vertical subset.
        models_library_path (Path): A path to the library file.

    Returns:
        np.ndarray: The generated 4D phantom.
    """
    if type(phantom_size) == tuple:
        N1, N2, N3 = [int(i) for i in phantom_size]
    else:
        N1 = N2 = N3 = phantom_size

    Z1, Z2 = [int(i) for i in sub_index]

    rangecheck = Z2 > Z1
    if not rangecheck:
        raise ValueError("Upper index must be larger than the lower one")
    rangecheck = Z1 >= 0 and Z1 < N3
    if not rangecheck:
        raise ValueError("Range of the lower index is incorrect")
    rangecheck = Z2 >= 0 and Z2 <= N3
    if not rangecheck:
        raise ValueError("Range of the higher index is incorrect")

    sub_size = Z2 - Z1  # the size of the vertical slab

    # check the validity of model's parameters
    params = np.ascontiguousarray(np.zeros([12], dtype=ctypes.c_int))
    params = _check_params3d(model_no, models_library_path)
    __testParams3D(params)  # check parameters and terminate if incorrect

    if params[3] == 1:
        raise ValueError(
            "The selected model is temporal (3D+time), use 'ModelTemporal' function instead"
        )
    else:
        phantom = np.zeros([params[3], sub_size, N2, N3], dtype="float32", order="C")
    external.c_model3d(
        np.ascontiguousarray(phantom),
        model_no,
        N1,
        N2,
        N3,
        Z1,
        Z2,
        models_library_path,
    )
    return phantom


def ModelSino(
    model_no: int,
    phantom_size: int,
    detector_horiz: int,
    detector_vert: int,
    angles: np.ndarray,
    models_library_path: Path,
) -> np.ndarray:
    """Generates a 3D analytical sinogram for corresponding models in
    the library file Phantom3DLibrary.dat.

    Args:
        model_no (int): Model number from the Phantom3DLibrary.dat library file.
        phantom_size (int): A scalar for squared phantom dimensions (2D slice).
        detector_horiz (int): Size of the horizontal detector in pixels.
        detector_vert (int): Size of the vertical detector in pixels.
        angles (np.ndarray): Angles vector in degrees.
        models_library_path (Path): A path to the library file.

    Returns:
        np.ndarray: The generated 3D projection data of the following shape [detector_vert, AngTot, detector_horiz].
    """
    if type(phantom_size) == tuple:
        raise ValueError(
            "Please give a scalar for the phantom size (2d slice), if one needs non-cubic data generator please see ModelSinoSub function"
        )

    # check the validity of model's parameters
    params = np.ascontiguousarray(np.zeros([12], dtype=ctypes.c_int))
    params = _check_params3d(model_no, models_library_path)
    __testParams3D(params)  # check parameters and terminate if incorrect

    angles_total = len(angles)
    if params[3] == 1:
        # execute the model building function
        sino3d = np.zeros(
            [angles_total, detector_vert, detector_horiz], dtype="float32", order="C"
        )
    else:
        raise ValueError(
            "The selected model is temporal (4D), use 'ModelTemporalSino' function instead"
        )
    external.c_model_sino3d(
        np.ascontiguousarray(sino3d),
        model_no,
        detector_horiz,
        detector_vert,
        0,
        detector_vert,
        phantom_size,
        np.ascontiguousarray(angles),
        angles_total,
        models_library_path,
    )
    return np.swapaxes(sino3d, 0, 1)


def ModelSinoSub(
    model_no: int,
    phantom_size: int,
    detector_horiz: int,
    detector_vert: int,
    sub_index: tuple,
    angles: np.ndarray,
    models_library_path: Path,
) -> np.ndarray:
    """Generates a 3D analytical sinogram with vertical cutoff using sub_index tuple for corresponding models in
    the library file Phantom3DLibrary.dat.

    Args:
        model_no (int): Model number from the Phantom3DLibrary.dat library file.
        phantom_size (int): A scalar for squared phantom dimensions (2D slice).
        detector_horiz (int): Size of the horizontal detector in pixels.
        detector_vert (int): Size of the vertical detector in pixels.
        sub_index (Tuple[int, int]): a tuple containing 2 indeces (lower, upper) to select a vertical subset.
        angles (np.ndarray): Angles vector in degrees.
        models_library_path (Path): A path to the library file.

    Returns:
        np.ndarray: The generated 3D projection data of the following shape [detector_vert, AngTot, detector_horiz].
    """
    if type(phantom_size) == tuple:
        raise ValueError(
            "Please give a scalar for phantom size, projection data cannot be obtained for non-cubic phantom"
        )

    Z1, Z2 = [int(i) for i in sub_index]

    rangecheck = Z2 > Z1
    if not rangecheck:
        raise ValueError("Upper index must be larger than the lower one")
    rangecheck = Z1 >= 0 and Z1 < detector_vert
    if not rangecheck:
        raise ValueError("Range of the lower index is incorrect")
    rangecheck = Z2 >= 0 and Z2 <= detector_vert
    if not rangecheck:
        raise ValueError("Range of the higher index is incorrect")

    sub_size = Z2 - Z1  # the size of the vertical slab

    # check the validity of model's parameters
    params = np.ascontiguousarray(np.zeros([12], dtype=ctypes.c_int))
    params = _check_params3d(model_no, models_library_path)
    __testParams3D(params)  # check parameters and terminate if incorrect

    angles_total = len(angles)
    if params[3] == 1:
        # execute the model building function
        sino3d = np.zeros(
            [angles_total, sub_size, detector_horiz], dtype="float32", order="C"
        )
    else:
        raise ValueError(
            "The selected model is temporal (4D), use 'ModelTemporalSino' function instead"
        )
    external.c_model_sino3d(
        np.ascontiguousarray(sino3d),
        model_no,
        detector_horiz,
        detector_vert,
        Z1,
        Z2,
        phantom_size,
        np.ascontiguousarray(angles),
        angles_total,
        models_library_path,
    )
    return np.swapaxes(sino3d, 0, 1)


def ModelSinoTemporal(
    model_no: int,
    phantom_size: int,
    detector_horiz: int,
    detector_vert: int,
    angles: np.ndarray,
    models_library_path: Path,
) -> np.ndarray:
    """Generates a 4D (3D+time) analytical sinogram for corresponding models in
    the library file Phantom3DLibrary.dat.

    Args:
        model_no (int): Temporal model number from the Phantom3DLibrary.dat library file.
        phantom_size (int): A scalar for squared phantom dimensions (2D slice).
        detector_horiz (int): Size of the horizontal detector in pixels.
        detector_vert (int): Size of the vertical detector in pixels.
        angles (np.ndarray): Angles vector in degrees.
        models_library_path (Path): A path to the library file.

    Returns:
        np.ndarray: The generated 4D projection data of the following shape [time_frames, detector_vert, AngTot, detector_horiz].
    """

    if type(phantom_size) == tuple:
        raise ValueError(
            "Please give a scalar for phantom size, projection data cannot be obtained for non-cubic phantom"
        )

    # check the validity of model's parameters
    params = np.ascontiguousarray(np.zeros([12], dtype=ctypes.c_int))
    params = _check_params3d(model_no, models_library_path)
    __testParams3D(params)  # check parameters and terminate if incorrect

    angles_total = len(angles)
    if params[3] == 1:
        raise ValueError(
            "The selected model is stationary (3D), use 'ModelSino' function instead"
        )
    else:
        sino4d = np.zeros(
            [params[3], angles_total, detector_vert, detector_horiz],
            dtype="float32",
            order="C",
        )

    external.c_model_sino3d(
        np.ascontiguousarray(sino4d),
        model_no,
        detector_horiz,
        detector_vert,
        0,
        detector_vert,
        phantom_size,
        np.ascontiguousarray(angles),
        angles_total,
        models_library_path,
    )
    return np.swapaxes(sino4d, 1, 2)


def ModelSinoTemporalSub(
    model_no: int,
    phantom_size: int,
    detector_horiz: int,
    detector_vert: int,
    sub_index: tuple,
    angles: np.ndarray,
    models_library_path: Path,
) -> np.ndarray:
    """Generates a 4D analytical sinogram with vertical cutoff using sub_index tuple for corresponding models in
    the library file Phantom3DLibrary.dat.

    Args:
        model_no (int): Temporal model number from the Phantom3DLibrary.dat library file.
        phantom_size (int): A scalar for squared phantom dimensions (2D slice).
        detector_horiz (int): Size of the horizontal detector in pixels.
        detector_vert (int): Size of the vertical detector in pixels.
        sub_index (Tuple[int, int]): a tuple containing 2 indeces (lower, upper) to select a vertical subset.
        angles (np.ndarray): Angles vector in degrees.
        models_library_path (Path): A path to the library file.

    Returns:
        np.ndarray: The generated 3D projection data of the following shape [time, detector_vert, AngTot, detector_horiz].
    """
    if type(phantom_size) == tuple:
        raise ValueError(
            "Please give a scalar for phantom size, projection data cannot be obtained for non-cubic phantom"
        )

    Z1, Z2 = [int(i) for i in sub_index]

    rangecheck = Z2 > Z1
    if not rangecheck:
        raise ValueError("Upper index must be larger than the lower one")
    rangecheck = Z1 >= 0 and Z1 < detector_vert
    if not rangecheck:
        raise ValueError("Range of the lower index is incorrect")
    rangecheck = Z2 >= 0 and Z2 <= detector_vert
    if not rangecheck:
        raise ValueError("Range of the higher index is incorrect")

    sub_size = Z2 - Z1  # the size of the vertical slab

    # check the validity of model's parameters
    params = np.ascontiguousarray(np.zeros([12], dtype=ctypes.c_int))
    params = _check_params3d(model_no, models_library_path)
    __testParams3D(params)  # check parameters and terminate if incorrect

    angles_total = len(angles)
    if params[3] == 1:
        raise ValueError(
            "The selected model is stationary (3D), use 'ModelSino' function instead"
        )
    else:
        sino4d = np.zeros(
            [params[3], angles_total, sub_size, detector_horiz],
            dtype="float32",
            order="C",
        )

    external.c_model_sino3d(
        np.ascontiguousarray(sino4d),
        model_no,
        detector_horiz,
        detector_vert,
        Z1,
        Z2,
        phantom_size,
        np.ascontiguousarray(angles),
        angles_total,
        models_library_path,
    )
    return np.swapaxes(sino4d, 1, 2)


def Object(
    phantom_size: Union[int, Tuple[int, int, int]], obj_params: Union[list, dict]
) -> np.ndarray:
    """Generates a 3D object for the standalone geometrical figure
    that is parametrised in the "obj_params" dictionary, see :ref:`ref_object_api`.

    Args:
        phantom_size (int, Tuple(int)): A scalar or a tuple with phantom dimensions.
        obj_params (a list of dicts or dict): A dictionary with parameters of an object, see :ref:`ref_object_api`.

    Returns:
        np.ndarray: The generated 3D object.
    """

    if type(phantom_size) == tuple:
        N1, N2, N3 = [int(i) for i in phantom_size]
    else:
        N1 = N2 = N3 = phantom_size

    if type(obj_params) is dict:
        obj_params = [obj_params]

    object3d = np.zeros([N1, N2, N3], dtype="float32", order="C")
    # unpacking obj_params dictionary
    for obj in obj_params:
        if __testParams3d(obj):
            objectName = obj["Obj"].value
            C0 = obj["C0"]
            x0 = obj["x0"]
            y0 = obj["y0"]
            z0 = obj["z0"]
            a = obj["a"]
            b = obj["b"]
            c = obj["c"]
            phi1 = obj["phi1"]
            phi2 = 0.0
            phi3 = 0.0
            tt = 0
            external.c_object3d(
                np.ascontiguousarray(object3d),
                N1,
                N2,
                N3,
                0,
                N1,
                objectName,
                C0,
                x0,
                y0,
                z0,
                a,
                b,
                c,
                phi1,
                phi2,
                phi3,
                tt,
            )
    return object3d


def ObjectSino(
    phantom_size: int,
    detector_horiz: int,
    detector_vert: int,
    angles: np.ndarray,
    obj_params: Union[list, dict],
) -> np.ndarray:
    """Generates a 3D analytical projection data for the standalone geometrical figure
    that is parametrised in the "obj_params" dictionary, see :ref:`ref_object_api`.

    Args:
        phantom_size (int): A scalar for phantom dimensions.
        detector_horiz (int): Size of the horizontal detector in pixels.
        detector_vert (int): Size of the vertical detector in pixels.
        angles (np.ndarray): Angles vector in degrees.
        obj_params (a list of dicts or dict): A dictionary with parameters of an object,  see :ref:`ref_object_api`.

    Returns:
        np.ndarray: The generated 3D projection data for an object.
    """
    if type(phantom_size) == tuple:
        raise ValueError(
            "Please give a scalar for phantom size, projection data cannot be obtained for non-cubic phantom"
        )

    if type(obj_params) is dict:
        obj_params = [obj_params]

    angles_total = len(angles)
    tt = 0
    sino3d = np.zeros(
        [angles_total, detector_vert, detector_horiz], dtype="float32", order="C"
    )
    # unpacking obj_params dictionary
    for obj in obj_params:
        if __testParams3d(obj):
            objectName = obj["Obj"].value
            C0 = obj["C0"]
            x0 = obj["x0"]
            y0 = obj["y0"]
            z0 = obj["z0"]
            a = obj["a"]
            b = obj["b"]
            c = obj["c"]
            phi1 = obj["phi1"]
            phi2 = 0.0
            phi3 = 0.0
            tt = 0
            external.c_object_sino3d(
                np.ascontiguousarray(sino3d),
                detector_horiz,
                detector_vert,
                0,
                detector_vert,
                phantom_size,
                angles,
                angles_total,
                objectName,
                C0,
                x0,
                y0,
                z0,
                a,
                b,
                c,
                phi1,
                phi2,
                phi3,
                tt,
            )
    return np.swapaxes(sino3d, 0, 1)


def __testParams3d(obj):
    """Performs a simple type check of the input parameters and a range check"""
    if not type(obj) is dict:
        raise TypeError("obj is not a dict {0}".format(type(obj)))
    # type check
    for k, v in obj.items():
        if not isinstance(v, Number):
            if not k == "Obj":
                raise TypeError(k, "is not a Number")
    typecheck = True

    # range check
    rangecheck = obj["x0"] >= -1 and obj["x0"] <= 1
    if not rangecheck:
        raise ValueError("x0 is out of range. Must be between -1 and 1")
    rangecheck = rangecheck and obj["y0"] >= -1 and obj["y0"] <= 1
    if not rangecheck:
        raise ValueError("y0 is out of range. Must be between -1 and 1")
    rangecheck = rangecheck and obj["z0"] >= -1 and obj["z0"] <= 1
    if not rangecheck:
        raise ValueError("z0 is out of range. Must be between -1 and 1")
    rangecheck = rangecheck and obj["a"] > 0 and obj["a"] <= 2
    if not rangecheck:
        raise ValueError("a (object size) must be positive in [0,2] range")
    rangecheck = rangecheck and obj["b"] > 0 and obj["b"] <= 2
    if not rangecheck:
        raise ValueError("b (object size) must be positive in [0,2] range")
    rangecheck = rangecheck and obj["c"] > 0 and obj["c"] <= 2
    if not rangecheck:
        raise ValueError("c (object size) must be positive in [0,2] range")
    return rangecheck and typecheck


def __testParams3D(obj):
    if obj[0] == 0:
        raise TypeError(
            "Check if the library file <Phantom3DLibrary.dat> exists, the given path is correct and the syntax is valid"
        )
    if obj[1] == 0:
        raise TypeError(
            "The given model is not found, check available models in <Phantom3DLibrary.dat> file"
        )
    if obj[2] == 0:
        raise TypeError(
            "Components number cannot be negative, check <Phantom3DLibrary.dat> file"
        )
    if obj[3] == 0:
        raise TypeError(
            "TimeSteps cannot be negative, check <Phantom3DLibrary.dat> file"
        )
    if obj[4] == 0:
        raise TypeError("Unknown name of the object, check <Phantom3DLibrary.dat> file")
    if obj[5] == 0:
        raise TypeError(
            "C0 should not be equal to zero, check <Phantom3DLibrary.dat> file"
        )
    if obj[6] == 0:
        raise TypeError(
            "x0 (object position) must be in [-1,1] range, check <Phantom3DLibrary.dat> file"
        )
    if obj[7] == 0:
        raise TypeError(
            "y0 (object position) must be in [-1,1] range, check <Phantom3DLibrary.dat> file"
        )
    if obj[8] == 0:
        raise TypeError(
            "z0 (object position) must be in [-1,1] range, check <Phantom3DLibrary.dat> file"
        )
    if obj[9] == 0:
        raise TypeError(
            "a (object size) must be positive in [0,2] range, check <Phantom3DLibrary.dat> file"
        )
    if obj[10] == 0:
        raise TypeError(
            "b (object size) must be positive in [0,2] range, check <Phantom3DLibrary.dat> file"
        )
    if obj[11] == 0:
        raise TypeError(
            "c (object size) must be positive in [0,2] range, check <Phantom3DLibrary.dat> file"
        )
    return 0



"""Modules for generating 2D synthetic phantoms and their analytical projection data.
The dynamic phantoms (resulting in 3D arrays) can also be generated.

API Summary:

* :func:`Model` - generates a 2D phantom using a :ref:`ref_glossary_model` from :ref:`howto_2d_libs`.
* :func:`ModelTemporal` - generates a :ref:`ref_glossary_dynamic_model` (2D+t) phantom from :ref:`howto_2d_libs`.
* :func:`ModelSino` - generates a 2D sinogram for a :ref:`ref_glossary_model` from :ref:`howto_2d_libs`.
* :func:`ModelSinoTemporal` - generates a :ref:`ref_glossary_dynamic_model` (2D+t) sinogram from :ref:`howto_2d_libs`.
* :func:`Object` - generates one elementary :ref:`ref_glossary_object` from the provided parameters, see :ref:`ref_object_api`.
* :func:`ObjectSino` - generates a 2D sinogram for one :ref:`ref_glossary_object` from the provided parameters, see :ref:`ref_object_api`.
* :func:`SinoNum` - calculates a parallel-beam sinogram numerically from any given 2D array.
"""

import ctypes
import numpy as np
from numbers import Number
from enum import Enum
from typing import Union
from pathlib import Path

import tomophantom.ctypes.external as external

__all__ = [
    "Model",
    "ModelTemporal",
    "ModelSino",
    "ModelSinoTemporal",
    "Object",
    "ObjectSino",
    "SinoNum",
]


class Objects2D(Enum):
    """Enumeration with the available objects for 2D phantoms"""

    GAUSSIAN = "gaussian"
    PARABOLA = "parabola"
    PARABOLA1 = "parabola1"
    ELLIPSE = "ellipse"
    CONE = "cone"
    RECTANGLE = "rectangle"


def _check_params2d(model_no: int, models_library_path: Path) -> np.ndarray:
    """Check parameters before executing the generation script.

    Args:
        model_no (int): Model number from the Phantom2DLibrary.dat library file.
        models_library_path (Path): A path to the library file.

    Returns:
        list: a list of integers
    """
    params = np.ascontiguousarray(np.zeros([10], dtype=ctypes.c_int))
    external.c_checkParams2D(
        params,
        model_no,
        models_library_path,
    )
    return params


def Model(model_no: int, phantom_size: int, models_library_path: Path) -> np.ndarray:
    """Generate 2D phantoms based on the model no. in
    the library file :ref:`howto_2d_libs`.

    Args:
        model_no (int): Model number from the :ref:`howto_2d_libs`.
        phantom_size (int): A size of the generated phantom (squared).
        models_library_path (Path): A path to the library file.

    Returns:
        np.ndarray: The generated 2D phantom (N x N).
    """
    # check the validity of model's parameters
    params = np.ascontiguousarray(np.zeros([10], dtype=ctypes.c_int))
    params = _check_params2d(model_no, models_library_path)
    __testParams2D(params)  # check parameters and terminate if incorrect

    if params[3] == 1:
        # execute the model building function
        phantom = np.zeros([phantom_size, phantom_size], dtype="float32", order="C")
    else:
        raise ValueError(
            "The selected model is temporal (2D+time), use 'ModelTemporal' function instead"
        )
    external.c_model2d(
        np.ascontiguousarray(phantom),
        model_no,
        phantom_size,
        models_library_path,
    )
    return phantom


def ModelTemporal(
    model_no: int, phantom_size: int, models_library_path: Path
) -> np.ndarray:
    """Generate 2D+time temporal phantoms based on the model no. in
    the :ref:`howto_2d_libs`. Note that temporal phantom
    numbers begin at 100 onwards.

    Args:
        model_no (int): Temporal model number from the :ref:`howto_2d_libs`.
        phantom_size (int): A size of the generated phantom (squared).
        models_library_path (Path): A path to the library file.

    Returns:
        np.ndarray: The generated 2D+time phantom (time_frames x N x N).
    """
    # check the validity of model's parameters
    params = np.ascontiguousarray(np.zeros([10], dtype=ctypes.c_int))
    params = _check_params2d(model_no, models_library_path)
    __testParams2D(params)  # check parameters and terminate if incorrect

    if params[3] == 1:
        raise ValueError(
            "The selected model is stationary (2D), use 'Model' function instead"
        )
    else:
        # execute the model building function
        phantom = np.zeros(
            [params[3], phantom_size, phantom_size], dtype="float32", order="C"
        )
    external.c_model2d(
        np.ascontiguousarray(phantom),
        model_no,
        phantom_size,
        models_library_path,
    )
    return phantom


def ModelSino(
    model_no: int,
    phantom_size: int,
    detector_size: int,
    angles: np.ndarray,
    models_library_path: Path,
) -> np.ndarray:
    """Generate 2D analytical sinogram for corresponding models in
    the :ref:`howto_2d_libs`.

    Args:
        model_no (int): Model number from the :ref:`howto_2d_libs`.
        phantom_size (int): A size of the phantom (squared).
        detector_size (int): A size of the horizontal detector.
        angles (np.ndarray): Angles vector in degrees.
        models_library_path (Path): A path to the library file.

    Returns:
        np.ndarray: The generated 2D analytical sinogram (angles_total x detector_size).
    """
    # check the validity of model's parameters
    params = np.ascontiguousarray(np.zeros([10], dtype=ctypes.c_int))
    params = _check_params2d(model_no, models_library_path)
    __testParams2D(params)  # check parameters and terminate if incorrect

    angles_total = len(angles)
    if params[3] == 1:
        # execute the model building function
        sino2d = np.zeros([detector_size, angles_total], dtype="float32", order="C")
    else:
        raise ValueError(
            "The selected model is temporal (2D+time), use 'ModelSinoTemporal' function instead"
        )
    external.c_model_sino2d(
        np.ascontiguousarray(sino2d),
        model_no,
        phantom_size,
        detector_size,
        np.ascontiguousarray(angles),
        angles_total,
        models_library_path,
    )
    return sino2d.transpose()


def ModelSinoTemporal(
    model_no: int,
    phantom_size: int,
    detector_size: int,
    angles: np.ndarray,
    models_library_path: Path,
) -> np.ndarray:
    """Generate 2D+time (temporal )analytical sinogram for
    corresponding models in the :ref:`howto_2d_libs`.

    Args:
        model_no (int): Temporal model number from the :ref:`howto_2d_libs`.
        phantom_size (int): A size of the phantom (squared).
        detector_size (int): A size of the horizontal detector.
        angles (np.ndarray): Angles vector in degrees.
        models_library_path (Path): A path to the library file.

    Returns:
        np.ndarray: The generated 2D+time analytical sinogram (time_frames x detector_size x angles_total).
    """
    # check the validity of model's parameters
    params = np.ascontiguousarray(np.zeros([10], dtype=ctypes.c_int))
    params = _check_params2d(model_no, models_library_path)
    __testParams2D(params)  # check parameters and terminate if incorrect

    angles_total = len(angles)
    if params[3] == 1:
        raise ValueError(
            "The selected model is stationary (2D), use 'ModelSino' function instead"
        )
    else:
        # execute the model building function
        sino2d_t = np.zeros(
            [params[3], detector_size, angles_total], dtype="float32", order="C"
        )
    external.c_model_sino2d(
        np.ascontiguousarray(sino2d_t),
        model_no,
        phantom_size,
        detector_size,
        np.ascontiguousarray(angles),
        angles_total,
        models_library_path,
    )
    return sino2d_t


def Object(phantom_size: int, obj_params: Union[list, dict]) -> np.ndarray:
    """Generates a 2D analytical phantom for the standalone
       geometrical object that is parametrised in the "obj_params" dictionary, see :ref:`ref_object_api`. Multiple objects can be stacked together
       by providing a list of dictionaries, where each dictionary associated with an object. See more in

    Args:
        phantom_size (int): A size of the phantom (squared).
        obj_params (a dict or a list of dict): A dictionary or a list of dictionaries with the parameters for object(s), see :ref:`ref_object_api`.

    Returns:
        np.ndarray: The generated 2D analytical phantom.
    """
    if type(obj_params) is dict:
        obj_params = [obj_params]

    object2d = np.zeros([phantom_size, phantom_size], dtype="float32", order="C")
    # unpacking obj_params dictionary
    for obj in obj_params:
        if __testParams(obj):
            objectName = obj["Obj"].value
            C0 = obj["C0"]
            x0 = obj["x0"]
            y0 = obj["y0"]
            a = obj["a"]
            b = obj["b"]
            phi = obj["phi"]
            external.c_object2d(
                np.ascontiguousarray(object2d),
                phantom_size,
                objectName,
                C0,
                x0,
                y0,
                a,
                b,
                phi_ang=phi,
                tt=0,
            )
    return object2d


def ObjectSino(
    phantom_size: int,
    detector_size: int,
    angles: np.ndarray,
    obj_params: Union[list, dict],
) -> np.ndarray:
    """Generate a 2D analytical sinogram for the standalone
       geometrical object that is parametrised in the "obj_params" dictionary.

    Args:
        phantom_size (int): A size of the phantom (squared).
        detector_size (int): A size of the horizontal detector.
        angles (np.ndarray): Angles vector in degrees.
        obj_params (list of dicts or a dict): A dictionary with parameters of an object, see :ref:`ref_object_api`.

    Returns:
        np.ndarray: The generated 2D analytical sinogram of an object.
    """
    angles_total = len(angles)
    sino2d = np.zeros([detector_size, angles_total], dtype="float32", order="C")
    # unpacking obj_params dictionary
    for obj in obj_params:
        if __testParams(obj):
            objectName = obj["Obj"].value
            C0 = obj["C0"]
            x0 = obj["x0"]
            y0 = obj["y0"]
            a = obj["a"]
            b = obj["b"]
            phi = obj["phi"]
            external.c_object_sino2d(
                np.ascontiguousarray(sino2d),
                phantom_size,
                detector_size,
                np.ascontiguousarray(angles),
                angles_total,
                objectName,
                C0,
                x0,
                y0,
                a,
                b,
                phi_ang=-phi,
                tt=0,
            )
    return sino2d.transpose()


def SinoNum(input: np.ndarray, detector_X: int, angles: np.ndarray) -> np.ndarray:
    """Numerical calculation of 2D sinogram from the 2D input.

    Args:
        input (np.ndarray): 2D object (e.g. a phantom).
        detector_X (int): The size of the detector X.
        angles (np.ndarray): Angles vector in degrees.

    Raises:
        ValueError: if input is 3D

    Returns:
        np.ndarray: Numerical sinogram calculated from the object
    """
    if np.ndim(input) == 3:
        raise ValueError("The accepted inputs must be 2D")
    phantom_size = input.shape[0]
    angles_total = len(angles)

    sinogram = np.zeros([detector_X, angles_total], dtype="float32", order="C")

    external.c_numerical_sino2d(
        np.ascontiguousarray(sinogram),
        np.ascontiguousarray(input),
        phantom_size,
        detector_X,
        np.ascontiguousarray(angles),
        angles_total,
    )
    return sinogram.transpose()


def __testParams(obj):
    """Performs a simple type check of the input parameters and a range check"""
    if not type(obj) is dict:
        raise TypeError("obj is not a dict {0}".format(type(obj)))

    # type check
    for k, v in obj.items():
        if not isinstance(v, Number):
            if not k == "Obj":
                raise TypeError(k, "is not a Number")
    typecheck = True
    # range check
    rangecheck = obj["x0"] >= -1 and obj["x0"] <= 1
    if not rangecheck:
        raise ValueError("x0 is out of range. Must be between -1 and 1")
    rangecheck = rangecheck and obj["y0"] >= -1 and obj["y0"] <= 1
    if not rangecheck:
        raise ValueError("y0 is out of range. Must be between -1 and 1")
    rangecheck = rangecheck and obj["a"] > 0 and obj["a"] <= 2
    if not rangecheck:
        raise ValueError("a (object size) must be positive in [0,2] range")
    rangecheck = rangecheck and obj["b"] > 0 and obj["b"] <= 2
    if not rangecheck:
        raise ValueError("b (object size) must be positive in [0,2] range")
    return rangecheck and typecheck


def __testParams2D(obj):
    if obj[0] == 0:
        raise TypeError(
            "Check if the library file <Phantom2DLibrary.dat> exists and the given path is correct"
        )
    if obj[1] == 0:
        raise TypeError(
            "The given model is not found, check available models in <Phantom2DLibrary.dat> file"
        )
    if obj[2] == 0:
        raise TypeError(
            "Components number cannot be negative, check <Phantom2DLibrary.dat> file"
        )
    if obj[3] == 0:
        raise TypeError(
            "TimeSteps cannot be negative, check <Phantom2DLibrary.dat> file"
        )
    if obj[4] == 0:
        raise TypeError("Unknown name of the object, check <Phantom2DLibrary.dat> file")
    if obj[5] == 0:
        raise TypeError(
            "C0 should not be equal to zero, check <Phantom2DLibrary.dat> file"
        )
    if obj[6] == 0:
        raise TypeError(
            "x0 (object position) must be in [-1,1] range, check <Phantom2DLibrary.dat> file"
        )
    if obj[7] == 0:
        raise TypeError(
            "y0 (object position) must be in [-1,1] range, check <Phantom2DLibrary.dat> file"
        )
    if obj[8] == 0:
        raise TypeError(
            "a (object size) must be positive in [0,2] range, check <Phantom2DLibrary.dat> file"
        )
    if obj[9] == 0:
        raise TypeError(
            "b (object size) must be positive in [0,2] range, check <Phantom2DLibrary.dat> file"
        )
    return 0



import numpy as np
from numpy.testing import assert_allclose

import pytest
import os
import tomophantom
from tomophantom.TomoP2D import Objects2D, ObjectSino

eps = 1e-05


def test_2d_object_sino():
    N_size = 64  # set the desired dimension of the phantom
    angles_num = int(0.5 * np.pi * N_size)
    # angles number
    angles = np.linspace(0.0, 179.9, angles_num, dtype="float32")
    P = int(np.sqrt(2) * N_size)  # detectors

    # define all objects bellow:
    pp = {
        "Obj": Objects2D.GAUSSIAN,
        "C0": 1.00,
        "x0": 0.25,
        "y0": -0.3,
        "a": 0.15,
        "b": 0.3,
        "phi": -30.0,
    }

    pp1 = {
        "Obj": Objects2D.RECTANGLE,
        "C0": 1.00,
        "x0": -0.2,
        "y0": 0.2,
        "a": 0.25,
        "b": 0.4,
        "phi": 60.0,
    }

    myObjects = [pp, pp1]  # list of dicts
    sino_an = ObjectSino(N_size, P, angles, myObjects)

    assert 0.0 <= np.max(sino_an) <= 1000
    assert sino_an.dtype == np.float32
    assert sino_an.shape == (angles_num, P)



"""Modules to generate synthetic flats/darks and data."""

from scipy.special import spherical_yn
from scipy.ndimage import gaussian_filter, shift
import random
import numpy as np
from tomophantom.artefacts import noise
from tomophantom.supp.speckle_routines import simulate_speckles_with_shot_noise


def synth_flats(
    projData3D_clean: np.ndarray,
    source_intensity: int,
    detectors_miscallibration: float = 0.05,
    variations_number: int = 3,
    arguments_Bessel: tuple = (1, 25),
    specklesize: int = 2,
    kbar: int = 2,
    sigmasmooth: int = 3,
    jitter_projections: float = 0.0,
    flatsnum: int = 20,
) -> list:
    """Function to generate synthetic flat field images and raw data for projection data normalisation.
    This is more realistic modelling of stripes and various artefacts rising after normalisation operation.

    Args:
        projData3D_clean (np.ndarray):  3D projection data of the following shape: (DetectorsVert, anglesDim, DetectorsHoriz).
        source_intensity (int): Source intensity which affects the amount of the Poisson noise added to data.
        detectors_miscallibration (float, optional): A constant which perturbs some detectors positions, leading to more prounonced ring artefacts. Defaults to 0.05.
        variations_number (int, optional): A type of function to control stripe type (1 - linear, 2, - sinusoidal, 3 - exponential). Defaults to 3.
        arguments_Bessel (tuple, optional): A tuple of 2 arguments for 2 Bessel functions to control background variations in flats. Defaults to (1,25).
        specklesize (int, optional):  Speckle size in pixel units for flats background simulation. Defaults to 2.
        kbar (int, optional): Mean photon density (photons per pixel) for background simulation. Defaults to 2.
        sigmasmooth (int, optional): Gaussian smoothing parameter to blur the speckled backround (1,3,5,7,...). Defaults to 3.
        jitter_projections (float, optional): An additional random jitter to the projections in pixels. Defaults to 0.0.
        flatsnum (int, optional): A number of flats to generate. Defaults to 20.

    Returns:
        list: List that contains [np.uint16(projData3D_raw), np.uint16(simulated_flats), blurred_speckles_map]
    """
    [DetectorsDimV, projectionsNo, DetectorsDimH] = np.shape(projData3D_clean)

    # output datasets
    flats_combined3D = np.zeros(
        (DetectorsDimV, flatsnum, DetectorsDimH), dtype="uint16"
    )
    projData3D_raw = np.zeros(np.shape(projData3D_clean), dtype="float32")

    # normalise the data
    projData3D_clean /= np.max(projData3D_clean)

    # using spherical Bessel functions to emulate the background (scintillator) variations
    func = spherical_yn(
        1,
        np.linspace(
            arguments_Bessel[0], arguments_Bessel[1], DetectorsDimV, dtype="float32"
        ),
    )
    func += abs(np.min(func))

    flatfield = np.zeros((DetectorsDimV, DetectorsDimH))
    for i in range(0, DetectorsDimH):
        flatfield[:, i] = func
    for i in range(0, DetectorsDimH):
        flatfield[:, i] += np.flipud(func)

    if specklesize != 0.0:
        # using speckle generator routines to create a photon count texture in the background
        speckle_background = simulate_speckles_with_shot_noise(
            [DetectorsDimV, DetectorsDimH], 1, specklesize, kbar
        )
    else:
        speckle_background = np.ones((DetectorsDimV, DetectorsDimH))

    # model miscallibrated detectors (a possible path to generate ring artifacts)
    blurred_speckles_map = np.zeros((DetectorsDimV, DetectorsDimH, variations_number))
    for i in range(0, variations_number):
        speckles = simulate_speckles_with_shot_noise(
            [DetectorsDimV, DetectorsDimH], 1, 10, 0.03
        )
        # blur the speckled background
        blurred_speckles = gaussian_filter(speckles.copy(), sigma=sigmasmooth)
        # threshold the result
        blurred_speckles[blurred_speckles < 0.6 * np.max(blurred_speckles)] = 0
        blurred_speckles_map[:, :, i] = blurred_speckles
    blurred_speckles_map /= np.max(blurred_speckles_map)

    sinusoidal_response = (
        np.sin(np.linspace(0, 1.5 * np.pi, projectionsNo))
        + np.random.random(projectionsNo) * 0.1
    )
    sinusoidal_response /= np.max(sinusoidal_response)
    exponential_response = (
        np.exp(np.linspace(0, np.pi, projectionsNo))
        + np.random.random(projectionsNo) * 0.1
    )
    exponential_response /= np.max(exponential_response)

    # prepeare flat fields
    for i in range(0, flatsnum):
        # add speckled background to the initial image with the Bessel background
        flatfield_combined = flatfield.copy() + 0.5 * (
            speckle_background / np.max(speckle_background)
        )
        flatfield_combined /= np.max(flatfield_combined)

        # adding Poisson noise to flat fields
        flatfield_poisson = noise(
            flatfield_combined * source_intensity, source_intensity, noisetype="Poisson"
        )
        flatfield_poisson /= np.max(flatfield_poisson)

        flats_combined3D[:, i, :] = np.uint16(flatfield_poisson * 65535)

    # convert synthetic projections to raw-data like projection ready for normalisation
    for i in range(0, projectionsNo):
        proj_exp = (
            np.exp(-projData3D_clean[:, i, :]) * source_intensity * flatfield_poisson
        )  # raw projection
        for j in range(0, variations_number):
            if j == 0:
                # adding a consistent offset for certain detectors
                proj_exp += (
                    blurred_speckles_map[:, :, j]
                    * detectors_miscallibration
                    * source_intensity
                )
            if j == 1:
                # adding a sinusoidal-like response offset for certain detectors
                proj_exp += (
                    sinusoidal_response[i]
                    * blurred_speckles_map[:, :, j]
                    * detectors_miscallibration
                    * source_intensity
                )
            if j == 2:
                # adding an exponential response offset for certain detectors
                proj_exp += (
                    exponential_response[i]
                    * blurred_speckles_map[:, :, j]
                    * detectors_miscallibration
                    * source_intensity
                )

        projection_poisson = noise(proj_exp, source_intensity, noisetype="Poisson")

        # apply jitter to projections
        if jitter_projections != 0.0:
            horiz_shift = random.uniform(
                -jitter_projections, jitter_projections
            )  # generate random directional shift
            vert_shift = random.uniform(
                -jitter_projections, jitter_projections
            )  # generate random directional shift
            projection_poisson = shift(
                projection_poisson.copy(), [vert_shift, horiz_shift], mode="reflect"
            )
        projData3D_raw[:, i, :] = projection_poisson

    projData3D_raw /= np.max(projData3D_raw)
    return [
        np.uint16(projData3D_raw * 65535),
        np.uint16(flats_combined3D),
        blurred_speckles_map,
    ]



"""Modules to generate different imaging artefacts."""

import numpy as np
import random
from typing import Union, Any


def artefacts_mix(data: np.ndarray, **artefacts_dict: Any) -> Union[np.ndarray, list]:
    """A module to generate and apply a mix of various typical imaging artefacts to add to the simulated data.
       One can build various dictionaries with keywords arguments specified
       bellow and pass it to the function as: `artefacts_mix(data, **noise_dict, **zingers, **etc)`.

       DISCLAIMER: Note that most of the features are experimental and do not always reflect the
       accurate modelling of the real imaging artefacts.

    Args:
        data (np.ndarray): 2D or 3D numpy array (sinogram or 3D projection data).
        The input data must be of the following shape: 2D sinogram (anglesDim, DetectorsHoriz),
        3D projection data (DetectorsVert, anglesDim, DetectorsHoriz).
        **artefacts_dict (dict): A dictionary with keyword arguments related to different artefact type.

    Keyword Args:
        noise_type (str): Define noise type as  'Poisson' or 'Gaussian'.
        noise_amplitude (int, float): Photon flux for Poisson or variance for Gaussian noise.
        noise_seed (int): Seeds for noise generator. 'None' defines random generation.
        noise_prelog (bool): Set to True if pre-log (raw) data required.
        zingers_percentage (float):  The amount of zingers (dead pixels, outliers) to be added to the data.
        zingers_modulus (int):  Modulus to control the amount of 4/6 pixel clusters to be added.
        stripes_percentage (float): The amount of stripes in the data (rings in reconstruction).
        stripes_maxthickness (int): Maxthickness defines the maximal thickness of a stripe.
        stripes_intensity (float): Controls the intensity levels of stripes.
        stripe_type (str): Stripe types can be 'partial' or 'full'.
        stripes_variability (float): Variability multiplier to incorporate the change of intensity in stripe.
        datashifts_maxamplitude_pixel (int): Controls pixel-wise (integer) misalignment in data.
        datashifts_maxamplitude_subpixel (float): Controls subpixel misalignment in data.
        pve_strength (int): The strength of partial volume effect, responsible for the limited resolution of a scanner.
        fresnel_dist_observation (int): The distance for observation for fresnel propagator.
        fresnel_scale_factor (float): Fresnel propagator scaling.
        fresnel_wavelenght (float): Fresnel propagator wavelength.
        verbose (bool): Make the output of modules verbose.

    Returns:
        np.ndarray: 2D/3D numpy array with artefacts applied to input data.
        [np.ndarray, shifts]: a list of 2D/3D numpy array and simulated shifts.
    """
    ####### VERBOSE ########
    if "verbose" not in artefacts_dict:
        verbose = True
    else:
        verbose = artefacts_dict["verbose"]
    ####### NOISE DICTIONARY########
    _noise_: dict = {}
    if "noise_type" not in artefacts_dict:
        _noise_["noise_type"] = None
    else:
        _noise_["noise_type"] = artefacts_dict["noise_type"]
    if "noise_amplitude" not in artefacts_dict:
        _noise_["noise_amplitude"] = 1e5
    else:
        _noise_["noise_amplitude"] = artefacts_dict["noise_amplitude"]
    if "noise_seed" not in artefacts_dict:
        _noise_["noise_seed"] = None
    else:
        _noise_["noise_seed"] = artefacts_dict["noise_seed"]
    if "noise_prelog" not in artefacts_dict:
        _noise_["noise_prelog"] = None
    else:
        _noise_["noise_prelog"] = artefacts_dict["noise_prelog"]
    ####### ZINGERS ########
    _zingers_: dict = {}
    if "zingers_percentage" not in artefacts_dict:
        _zingers_["zingers_percentage"] = None
    else:
        _zingers_["zingers_percentage"] = artefacts_dict["zingers_percentage"]
    if "zingers_modulus" not in artefacts_dict:
        _zingers_["zingers_modulus"] = 10
    else:
        _zingers_["zingers_modulus"] = artefacts_dict["zingers_modulus"]
    ####### STRIPES ########
    _stripes_: dict = {}
    if "stripes_percentage" not in artefacts_dict:
        _stripes_["stripes_percentage"] = None
    else:
        _stripes_["stripes_percentage"] = artefacts_dict["stripes_percentage"]
    if "stripes_maxthickness" not in artefacts_dict:
        _stripes_["stripes_maxthickness"] = 1.0
    else:
        _stripes_["stripes_maxthickness"] = artefacts_dict["stripes_maxthickness"]
    if "stripes_intensity" not in artefacts_dict:
        _stripes_["stripes_intensity"] = 0.1
    else:
        _stripes_["stripes_intensity"] = artefacts_dict["stripes_intensity"]
    if "stripes_type" not in artefacts_dict:
        _stripes_["stripes_type"] = "full"
    else:
        _stripes_["stripes_type"] = artefacts_dict["stripes_type"]
    if "stripes_variability" not in artefacts_dict:
        _stripes_["stripes_variability"] = 0.0
    else:
        _stripes_["stripes_variability"] = artefacts_dict["stripes_variability"]
    ####### DATASHIFTS ########
    _datashifts_: dict = {}
    if "datashifts_maxamplitude_pixel" not in artefacts_dict:
        _datashifts_["datashifts_maxamplitude_pixel"] = None
    else:
        _datashifts_["datashifts_maxamplitude_pixel"] = artefacts_dict[
            "datashifts_maxamplitude_pixel"
        ]
    if "datashifts_maxamplitude_subpixel" not in artefacts_dict:
        _datashifts_["datashifts_maxamplitude_subpixel"] = None
    else:
        _datashifts_["datashifts_maxamplitude_subpixel"] = artefacts_dict[
            "datashifts_maxamplitude_subpixel"
        ]
    ####### PVE ########
    _pve_: dict = {}
    if "pve_strength" not in artefacts_dict:
        _pve_["pve_strength"] = None
    else:
        _pve_["pve_strength"] = artefacts_dict["pve_strength"]
    _fresnel_propagator_: dict = {}
    if "fresnel_dist_observation" not in artefacts_dict:
        _fresnel_propagator_["fresnel_dist_observation"] = None
    else:
        _fresnel_propagator_["fresnel_dist_observation"] = artefacts_dict[
            "fresnel_dist_observation"
        ]
    if "fresnel_scale_factor" not in artefacts_dict:
        _fresnel_propagator_["fresnel_scale_factor"] = 10
    else:
        _fresnel_propagator_["fresnel_scale_factor"] = artefacts_dict[
            "fresnel_scale_factor"
        ]
    if "fresnel_wavelenght" not in artefacts_dict:
        _fresnel_propagator_["fresnel_wavelenght"] = 0.0001
    else:
        _fresnel_propagator_["fresnel_wavelenght"] = artefacts_dict[
            "fresnel_wavelenght"
        ]
    ###########################################################################
    ################Applying artefacts and noise to the data###################
    ###########################################################################
    # PARTIAL VOLUME EFFECT
    if _pve_["pve_strength"] is not None:
        sino_artifacts = np.float32(pve(data=data, pve_strength=_pve_["pve_strength"]))
        if verbose is True:
            print("Partial volume effect (PVE) has been simulated.")
    else:
        sino_artifacts = np.float32(data)
    # FRESNEL PROPAGATOR
    if _fresnel_propagator_["fresnel_dist_observation"] is not None:
        sino_artifacts = np.float32(
            fresnel_propagator(
                data=sino_artifacts,
                dist_observation=_fresnel_propagator_["fresnel_dist_observation"],
                scale_factor=_fresnel_propagator_["fresnel_scale_factor"],
                wavelenght=_fresnel_propagator_["fresnel_wavelenght"],
            )
        )
        if verbose is True:
            print("Fresnel propagator has been simulated.")
    # ZINGERS
    if _zingers_["zingers_percentage"] is not None:
        sino_artifacts = np.float32(
            zingers(
                data=sino_artifacts,
                percentage=_zingers_["zingers_percentage"],
                modulus=_zingers_["zingers_modulus"],
            )
        )
        if verbose is True:
            print("Zingers have been added to the data.")
    # STRIPES
    if _stripes_["stripes_percentage"] is not None:
        sino_artifacts = np.float32(
            stripes(
                data=sino_artifacts,
                percentage=_stripes_["stripes_percentage"],
                maxthickness=_stripes_["stripes_maxthickness"],
                intensity_thresh=_stripes_["stripes_intensity"],
                stripe_type=_stripes_["stripes_type"],
                variability=_stripes_["stripes_variability"],
            )
        )
        if verbose is True:
            print("Stripes leading to ring artefacts have been simulated.")
    # DATASHIFTS
    if _datashifts_["datashifts_maxamplitude_pixel"] is not None:
        [sino_artifacts, shifts] = datashifts(
            data=sino_artifacts,
            maxamplitude=_datashifts_["datashifts_maxamplitude_pixel"],
        )
        if verbose is True:
            print("Data shifts have been simulated.")
    if _datashifts_["datashifts_maxamplitude_subpixel"] is not None:
        [sino_artifacts, shifts] = datashifts_subpixel(
            data=sino_artifacts,
            maxamplitude=_datashifts_["datashifts_maxamplitude_subpixel"],
        )
        if verbose is True:
            print("Data shifts (in subpixel precision) have been simulated.")
    # NOISE
    if _noise_["noise_type"] is not None:
        sino_artifacts = noise(
            data=sino_artifacts,
            sigma=_noise_["noise_amplitude"],
            noisetype=_noise_["noise_type"],
            seed=_noise_["noise_seed"],
            prelog=_noise_["noise_prelog"],
        )
        if verbose is True:
            print("{} noise has been added to the data.".format(_noise_["noise_type"]))

    if (_datashifts_["datashifts_maxamplitude_pixel"]) or (
        _datashifts_["datashifts_maxamplitude_subpixel"]
    ) is not None:
        return [np.float32(sino_artifacts), shifts]
    else:
        return np.float32(sino_artifacts)


def stripes(
    data: np.ndarray,
    percentage: float,
    maxthickness: int,
    intensity_thresh: float,
    stripe_type: str,
    variability: float,
) -> np.ndarray:
    """Function to add stripes (constant offsets) to sinograms or 3D projection data which results in rings in the
    reconstructed image.

    Args:
        data (np.ndarray): 2D sinogram (anglesDim, DetectorsHoriz) or 3D projection data (DetectorsVert, anglesDim, DetectorsHoriz).
        percentage (float): Percentage defines the amount of stripes in the data.
        maxthickness (int): Defines the maximal thickness of a stripe.
        intensity_thresh (float): Controls the intensity levels of stripes.
        stripe_type (str): Choose between 'partial' or 'full'.
        variability (float): Variability multiplier to incorporate change of intensity in the stripe.

    Raises:
        ValueError: Percentage is out of range.
        ValueError: Thickness is out of range.

    Returns:
        np.ndarray: 2D sinogram or 3D projection data with stripes
    """
    if data.ndim == 2:
        (anglesDim, DetectorsDimH) = np.shape(data)
    else:
        (DetectorsDimV, anglesDim, DetectorsDimH) = np.shape(data)
    if 0 < percentage <= 100:
        pass
    else:
        raise ValueError("percentage must be larger than zero but smaller than 100")
    if 0 <= maxthickness <= 10:
        pass
    else:
        raise ValueError("maximum thickness must be in [0,10] range")
    if stripe_type != "partial":
        stripe_type = "full"
    sino_stripes = data.copy()
    max_intensity = np.max(sino_stripes)
    range_detect = int((np.float32(DetectorsDimH)) * (np.float32(percentage) / 100.0))
    if data.ndim == 2:
        for x in range(range_detect):
            for mm in range(0, 20):
                randind = random.randint(0, DetectorsDimH - 1)  # generate random index
                if sino_stripes[0, randind] != 0.0:
                    break
            if stripe_type == "partial":
                randind_ang1 = random.randint(0, anglesDim)
                randind_ang2 = random.randint(0, anglesDim)
            else:
                randind_ang1 = 0
                randind_ang2 = anglesDim
            randthickness = random.randint(0, maxthickness)  # generate random thickness
            randintens = random.uniform(-1.0, 0.5)  # generate random multiplier
            intensity = max_intensity * randintens * intensity_thresh
            if (randind > 0 + randthickness) & (
                randind < DetectorsDimH - randthickness
            ):
                for x1 in range(-randthickness, randthickness + 1):
                    if variability != 0.0:
                        intensity_off = variability * max_intensity
                    else:
                        intensity_off = 0.0
                    for ll in range(randind_ang1, randind_ang2):
                        sino_stripes[ll, randind + x1] += intensity + intensity_off
                        intensity_off += (
                            [-1, 1][random.randrange(2)] * variability * max_intensity
                        )
                        # sino_stripes[randind_ang1:randind_ang2,randind+x1] += intensity
    else:
        for j in range(DetectorsDimV):
            for x in range(range_detect):
                for mm in range(0, 20):
                    randind = random.randint(
                        0, DetectorsDimH - 1
                    )  # generate random index
                    if sino_stripes[j, 0, randind] != 0.0:
                        break
                if stripe_type == "partial":
                    randind_ang1 = random.randint(0, anglesDim)
                    randind_ang2 = random.randint(0, anglesDim)
                else:
                    randind_ang1 = 0
                    randind_ang2 = anglesDim
                randthickness = random.randint(
                    0, maxthickness
                )  # generate random thickness
                randintens = random.uniform(-1, 0.5)  # generate random multiplier
                intensity = max_intensity * randintens * intensity_thresh
                if (randind > 0 + randthickness) & (
                    randind < DetectorsDimH - randthickness
                ):
                    for x1 in range(-randthickness, randthickness + 1):
                        if variability != 0.0:
                            intensity_off = variability * max_intensity
                        else:
                            intensity_off = 0.0
                        for ll in range(randind_ang1, randind_ang2):
                            sino_stripes[j, ll, randind + x1] += (
                                intensity + intensity_off
                            )
                            intensity_off += (
                                [-1, 1][random.randrange(2)]
                                * variability
                                * max_intensity
                            )
    return sino_stripes


def zingers(data: np.ndarray, percentage: float, modulus: int) -> np.ndarray:
    """Adding zingers (zero single pixels or small 4 pixels clusters)
     to sinograms or 6 voxels to 3D projection data.

    Args:
        data (np.ndarray): 2D sinogram or 3D projection data. The input data must be of the following shape: 2D sinogram (anglesDim, DetectorsHoriz),
        3D projection data (DetectorsVert, anglesDim, DetectorsHoriz).
        percentage (float): The amount of zingers to be added to the data.
        modulus (int): Modulus to control the amount of 4/6 pixel clusters to be added.

    Raises:
        ValueError: Percentage is out of range.
        ValueError: Modulus must be positive.

    Returns:
        np.ndarray: 2D or 3D array with zingers.
    """
    if data.ndim == 2:
        (anglesDim, DetectorsDimH) = np.shape(data)
    else:
        (DetectorsDimV, anglesDim, DetectorsDimH) = np.shape(data)
    if 0.0 < percentage <= 100.0:
        pass
    else:
        raise ValueError("percentage must be larger than zero but smaller than 100")
    if modulus > 0:
        pass
    else:
        raise ValueError("Modulus integer must be positive")
    sino_zingers = data.copy()
    length_sino = np.size(sino_zingers)
    num_values = int((length_sino) * (np.float32(percentage) / 100.0))
    sino_zingers_fl = sino_zingers.flatten()
    for x in range(num_values):
        randind = random.randint(0, length_sino - 1)  # generate random index
        sino_zingers_fl[randind] = 0
        if (x % int(modulus)) == 0:
            if data.ndim == 2:
                if (randind > DetectorsDimH) & (randind < length_sino - DetectorsDimH):
                    sino_zingers_fl[randind + 1] = 0
                    sino_zingers_fl[randind - 1] = 0
                    sino_zingers_fl[randind + DetectorsDimH] = 0
                    sino_zingers_fl[randind - DetectorsDimH] = 0
            else:
                if (randind > DetectorsDimH * DetectorsDimV) & (
                    randind < length_sino - DetectorsDimH * DetectorsDimV
                ):
                    sino_zingers_fl[randind + 1] = 0
                    sino_zingers_fl[randind - 1] = 0
                    sino_zingers_fl[randind + DetectorsDimH] = 0
                    sino_zingers_fl[randind - DetectorsDimH] = 0
                    sino_zingers_fl[randind + DetectorsDimH * DetectorsDimV] = 0
                    sino_zingers_fl[randind - DetectorsDimH * DetectorsDimV] = 0
    sino_zingers[:] = sino_zingers_fl.reshape(sino_zingers.shape)
    return sino_zingers


def noise(
    data: np.ndarray,
    sigma: Union[int, float],
    noisetype: str,
    seed: bool = True,
    prelog: bool = False,
) -> Union[list, np.ndarray]:
    """Adding random noise to data (adapted from LD-CT simulator)

    Args:
        data (np.ndarray): N-d array
        sigma (Union[int, float]): int for Poisson or float for Gaussian
        noisetype (str): 'Gaussian' or 'Poisson'
        seed (bool, optional): Initiate random seed with True. Defaults to True.
        prelog (bool, optional): get raw data if true, returns a list [noisy, raw]!. Defaults to False.

    Returns:
        np.ndarray: N-d array
    """
    data_noisy = data.copy()
    maxData = np.max(data)
    data_noisy = data / maxData  # normalising
    if seed is True:
        np.random.seed(int(seed))
    if noisetype == "Gaussian":
        # add normal Gaussian noise
        data_noisy += np.random.normal(loc=0.0, scale=sigma, size=np.shape(data_noisy))
        data_noisy[data_noisy < 0] = 0
        data_noisy *= maxData
    elif noisetype == "Poisson":
        # add Poisson noise
        if maxData > 0:
            ri = 1.0
            sig = np.sqrt(
                11
            )  # standard variance of electronic noise, a characteristic of a CT scanner
            yb = (
                sigma * np.exp(-data_noisy) + ri
            )  # exponential transform to incident flux
            data_raw = np.random.poisson(yb) + np.sqrt(sig) * np.reshape(
                np.random.randn(np.size(yb)), np.shape(yb)
            )
            li_hat = -np.log((data_raw - ri) / sigma) * maxData  # log corrected data
            li_hat[(data_raw - ri) <= 0] = 0.0
            data_noisy = li_hat.copy()
    else:
        print("Select 'Gaussian' or 'Poisson' for the noise type")
    if prelog is True:
        return [data_noisy, data_raw]
    else:
        return data_noisy


def datashifts(data: np.ndarray, maxamplitude: int) -> list:
    """Function to add random pixel shifts to 3D or 3D data as an offset for each angular position.

    Args:
        data (np.ndarray): 2D sinogram or 3D projection data. The input data must be of the following shape: 2D sinogram (anglesDim, DetectorsHoriz),
        3D projection data (DetectorsVert, anglesDim, DetectorsHoriz).
        maxamplitude (int): The misilighnment ammplitude. Defines the maximal amplitude of each shift in pixels.

    Returns:
        list: 2D or 3d data with misalignment and shifts vectors [data, shifts].
    """
    if data.ndim == 2:
        (anglesDim, DetectorsDimH) = np.shape(data)
        shifts = np.zeros(anglesDim, dtype="int8")  # the vector of shifts
    else:
        (DetectorsDimV, anglesDim, DetectorsDimH) = np.shape(data)
        shifts = np.zeros([anglesDim, 2], dtype="int8")  # the 2D vector of shifts

    sino_shifts = np.zeros(np.shape(data), dtype="float32")
    non = lambda s: s if s < 0 else None
    mom = lambda s: max(0, s)
    for x in range(anglesDim):
        rand_shift = random.randint(
            -maxamplitude, maxamplitude
        )  # generate random shift (int)
        if data.ndim == 2:
            shifts[x] = rand_shift
            projection = data[x, :]  # extract 1D projection
            projection_shift = np.zeros(np.shape(projection), dtype="float32")
            projection_shift[mom(rand_shift) : non(rand_shift)] = projection[
                mom(-rand_shift) : non(-rand_shift)
            ]
            sino_shifts[x, :] = projection_shift
        else:
            rand_shift2 = random.randint(
                -maxamplitude, maxamplitude
            )  # generate random shift (int)
            shifts[x, 0] = rand_shift2
            shifts[x, 1] = rand_shift
            projection2D = data[:, x, :]  # extract 2D projection
            projection2D_shift = np.zeros(np.shape(projection2D), dtype="float32")
            projection2D_shift[
                mom(rand_shift) : non(rand_shift), mom(rand_shift2) : non(rand_shift2)
            ] = projection2D[
                mom(-rand_shift) : non(-rand_shift),
                mom(-rand_shift2) : non(-rand_shift2),
            ]
            sino_shifts[:, x, :] = projection2D_shift
    return [sino_shifts, shifts]


def datashifts_subpixel(data: np.ndarray, maxamplitude: float) -> list:
    """Function to add random sub-pixel shifts to 3D or 3D data as an offset for each angular position.

    Args:
        data (np.ndarray): 2D sinogram or 3D projection data. The input data must be of the following shape: 2D sinogram (anglesDim, DetectorsHoriz),
        3D projection data (DetectorsVert, anglesDim, DetectorsHoriz).
        maxamplitude (float): The misilighnment ammplitude. Defines the maximal amplitude of each shift.

    Returns:
        list: 2D or 3d data with misalignment and shifts vectors [data, shifts].
    """
    from skimage import transform as tf

    if data.ndim == 2:
        shifts = np.zeros([1, 2], dtype="float32")
        random_shift_x = random.uniform(
            -maxamplitude, maxamplitude
        )  # generate a random floating point number
        random_shift_y = random.uniform(
            -maxamplitude, maxamplitude
        )  # generate a random floating point number

        tform = tf.SimilarityTransform(translation=(-random_shift_x, -random_shift_y))
        sino_shifts = tf.warp(data, tform, order=5)
    else:
        (DetectorsDimV, anglesDim, DetectorsDimH) = np.shape(data)
        shifts = np.zeros([anglesDim, 2], dtype="float32")  # the 2D vector of shifts
        sino_shifts = np.zeros(np.shape(data), dtype="float32")
        for x in range(anglesDim):
            random_shift_x = random.uniform(
                -maxamplitude, maxamplitude
            )  # generate a random floating point number
            random_shift_y = random.uniform(
                -maxamplitude, maxamplitude
            )  # generate a random floating point number

            projection2D = data[:, x, :]  # extract 2D projection
            tform = tf.SimilarityTransform(
                translation=(-random_shift_x, -random_shift_y)
            )
            projection_shifted = tf.warp(projection2D, tform, order=5)

            shifts[x, 0] = random_shift_x
            shifts[x, 1] = random_shift_y

            sino_shifts[:, x, :] = projection_shifted
    return [sino_shifts, shifts]


def pve(data: np.ndarray, pve_strength: int) -> np.ndarray:
    """Applying Partial Volume effect (smoothing) to data.

    Args:
        data (np.ndarray): 2D sinogram or 3D projection data. The input data must be of the following shape: 2D sinogram (anglesDim, DetectorsHoriz),
        3D projection data (DetectorsVert, anglesDim, DetectorsHoriz).
        pve_strength (int): The level of smoothing, defined by kernel size.

    Raises:
        ValueError: Smoothing kernel must be positive

    Returns:
        np.ndarray: Smoothed 2D or 3D data.
    """

    from scipy.ndimage import gaussian_filter

    data_pve = data.copy()
    if pve_strength > 0:
        pass
    else:
        raise ValueError("Smoothing kernel must be positive")
    if data.ndim == 2:
        (anglesDim, DetectorsDimH) = np.shape(data)
        for x in range(anglesDim):
            data_pve[x, :] = gaussian_filter(data_pve[x, :], pve_strength)
    else:
        (DetectorsDimV, anglesDim, DetectorsDimH) = np.shape(data)
        for x in range(anglesDim):
            data_pve[:, x, :] = gaussian_filter(data_pve[:, x, :], pve_strength)
    return data_pve


def fresnel_propagator(
    data: np.ndarray, dist_observation: int, scale_factor: float, wavelenght: float
) -> np.ndarray:
    """Fresnel propagator applied to data.
       Adapted from the script by Adrián Carbajal-Domínguez, adrian.carbajal@ujat.mx

    Args:
        data (np.ndarray): 2D sinogram or 3D projection data. The input data must be of the following shape: 2D sinogram (anglesDim, DetectorsHoriz),
        3D projection data (DetectorsVert, anglesDim, DetectorsHoriz).
        dist_observation (int): The distance for obervation for fresnel propagator.
        scale_factor (float): Scaling.
        wavelenght (float): Wavelength.

    Returns:
        np.ndarray: 2D or 3D data with propagation.
    """

    data_fresnel = data.copy()
    if data.ndim == 2:
        (anglesDim, DetectorsDimH) = np.shape(data)
        n1 = DetectorsDimH * 0.5
        # Define the angular spectrum coordinates
        u = np.arange(-n1, n1, 1)
        # Define the propagation matrix
        propagator = np.exp(
            2
            * np.pi
            * 1j
            * (dist_observation / scale_factor)
            * np.sqrt((1 / wavelenght) ** 2 - (u / 10) ** 2)
        )
        #### Compute the Fast Fourier Transform of each 1D projection
        for x in range(anglesDim):
            f = np.fft.fft(data_fresnel[x, :])
            # Correct the low and high frequencies
            fshift = np.fft.fftshift(f)
            # multiply both matrices: Fourier transform and the propagator matrices.
            field = fshift * propagator
            # Calculate the inverse Fourier transform
            field2 = np.fft.ifft(field)
            data_fresnel[x, :] = np.abs(field2)
    else:
        (DetectorsDimV, anglesDim, DetectorsDimH) = np.shape(data)
        ####Define the size of the propagation function p(u,v). It has to be of the same size of the image.
        n1 = DetectorsDimV * 0.5
        n2 = DetectorsDimH * 0.5
        # Define the angular spectrum coordinates
        u = np.arange(-n1, n1, 1)
        v = np.arange(-n2, n2, 1)
        U, V = np.meshgrid(u, v)
        # Define the propagation matrix
        propagator = np.exp(
            2
            * np.pi
            * 1j
            * (dist_observation / scale_factor)
            * np.sqrt(
                (1 / wavelenght) ** 2
                - (U / scale_factor) ** 2
                - (V / scale_factor) ** 2
            )
        )
        #### Compute the Fast Fourier Transform of each 2D projection
        for x in range(anglesDim):
            f = np.fft.fft2(data_fresnel[:, x, :])
            # Correct the low and high frequencies
            fshift = np.fft.fftshift(f)
            # multiply both matrices: Fourier transform and the propagator matrices.
            field = fshift * np.transpose(propagator)
            # Calculate the inverse Fourier transform
            field2 = np.fft.ifft2(field)
            data_fresnel[:, x, :] = np.abs(field2)
    return data_fresnel



"""The set of modules to generate 2D/3D :ref:`ref_glossary_random_phantoms`. Specifically "foam"-like phantoms."""

import numpy as np
import math
import random


def _rand_init2D(x0min, x0max, y0min, y0max, c0min, c0max, ab_min, ab_max):
    import numpy as np

    x0 = np.random.uniform(low=x0min, high=x0max)
    y0 = np.random.uniform(low=y0min, high=y0max)
    c0 = np.random.uniform(low=c0min, high=c0max)
    ab = np.random.uniform(low=ab_min, high=ab_max)
    return (x0, y0, c0, ab)


def _rand_init3D(
    x0min, x0max, y0min, y0max, z0min, z0max, c0min, c0max, ab_min, ab_max
):
    import numpy as np

    x0 = np.random.uniform(low=x0min, high=x0max)
    y0 = np.random.uniform(low=y0min, high=y0max)
    z0 = np.random.uniform(low=z0min, high=z0max)
    c0 = np.random.uniform(low=c0min, high=c0max)
    ab = np.random.uniform(low=ab_min, high=ab_max)
    return (x0, y0, z0, c0, ab)


def foam2D(
    x0min,
    x0max,
    y0min,
    y0max,
    c0min,
    c0max,
    ab_min,
    ab_max,
    N_size,
    tot_objects,
    object_type,
):
    """
    Function to generate 2D foam-like structures using different object types (ellipse, parabola, gaussian) with random parameters.
    """
    # 2D functions
    from tomophantom import TomoP2D
    from tomophantom.TomoP2D import Objects2D

    attemptsNo = 2000  # the number of attempts to fit the object
    # objects accepted: 'ellipse', 'parabola', 'gaussian', 'mix'
    mix_objects = False
    if object_type == "ellipse":
        object_type = Objects2D.ELLIPSE
    elif object_type == "parabola":
        object_type = Objects2D.PARABOLA
    elif object_type == "gaussian":
        object_type = Objects2D.GAUSSIAN
    elif object_type == "mix":
        mix_objects = True
    else:
        raise TypeError("object_type can be only ellipse, parabola, gaussian or mix")
    X0 = np.float32(np.zeros(tot_objects))
    Y0 = np.float32(np.zeros(tot_objects))
    AB = np.float32(np.zeros(tot_objects))
    C0_var = np.float32(np.zeros(tot_objects))
    for i in range(0, tot_objects):
        (x0, y0, c0, ab) = _rand_init2D(
            x0min, x0max, y0min, y0max, c0min, c0max, ab_min, ab_max
        )
        if i > 0:
            breakj = False
            for j in range(0, attemptsNo):
                if breakj == True:
                    (x0, y0, c0, ab) = _rand_init2D(
                        x0min, x0max, y0min, y0max, c0min, c0max, ab_min, ab_max
                    )
                    breakj = False
                else:
                    for l in range(
                        0, i
                    ):  # checks consistency with previously created objects
                        dist = math.sqrt((X0[l] - x0) ** 2 + (Y0[l] - y0) ** 2)
                        if (dist < (ab + AB[l])) or (
                            (abs(x0) + ab) ** 2 + (abs(y0) + ab) ** 2 > 1.0
                        ):
                            breakj = True
                            break
                    if breakj == False:  # re-initialise if doesn't fit the criteria
                        X0[i] = x0
                        Y0[i] = y0
                        AB[i] = ab
                        C0_var[i] = c0
                        break
        if AB[i] == 0.0:
            X0[i] = x0
            Y0[i] = y0
            AB[i] = 0.0001
            C0_var[i] = c0

    myObjects = []  # dictionary of objects
    for obj in range(0, len(X0)):
        if mix_objects == True:
            rand_obj = random.randint(0, 2)
            if rand_obj == 0:
                object_type = Objects2D.ELLIPSE
            if rand_obj == 1:
                object_type = Objects2D.PARABOLA
            if rand_obj == 2:
                object_type = Objects2D.GAUSSIAN
        curr_obj = {
            "Obj": object_type,
            "C0": C0_var[obj],
            "x0": X0[obj],
            "y0": Y0[obj],
            "a": AB[obj],
            "b": AB[obj],
            "phi": 0.0,
        }
        myObjects.append(curr_obj)
    Object = TomoP2D.Object(N_size, myObjects)
    return (Object, myObjects)


# Function to generate 3D foam-like structures using randomly located spheres
def foam3D(
    x0min,
    x0max,
    y0min,
    y0max,
    z0min,
    z0max,
    c0min,
    c0max,
    ab_min,
    ab_max,
    N_size,
    tot_objects,
    object_type,
):
    """
    Function to generate 3D foam-like structures using different object types (ellipsoids, paraboloids, gaussians) with random parameters.
    """
    # 3D functions
    from tomophantom import TomoP3D
    from tomophantom.TomoP3D import Objects3D

    attemptsNo = 2000
    # objects accepted: 'ellipsoid', 'paraboloid', 'gaussian', 'mix'
    mix_objects = False
    if object_type == "ellipsoid":
        object_type = Objects3D.ELLIPSOID
    elif object_type == "paraboloid":
        object_type = Objects3D.PARABOLOID
    elif object_type == "gaussian":
        object_type = Objects3D.GAUSSIAN
    elif object_type == "mix":
        mix_objects = True
    else:
        raise TypeError("object_type can be only ellipse, parabola, gaussian or mix")
    X0 = np.float32(np.zeros(tot_objects))
    Y0 = np.float32(np.zeros(tot_objects))
    Z0 = np.float32(np.zeros(tot_objects))
    AB = np.float32(np.zeros(tot_objects))
    C0_var = np.float32(np.zeros(tot_objects))
    for i in range(0, tot_objects):
        (x0, y0, z0, c0, ab) = _rand_init3D(
            x0min, x0max, y0min, y0max, z0min, z0max, c0min, c0max, ab_min, ab_max
        )
        if i > 0:
            breakj = False
            for j in range(0, attemptsNo):
                if breakj:
                    (x0, y0, z0, c0, ab) = _rand_init3D(
                        x0min,
                        x0max,
                        y0min,
                        y0max,
                        z0min,
                        z0max,
                        c0min,
                        c0max,
                        ab_min,
                        ab_max,
                    )
                    breakj = False
                else:
                    for l in range(
                        0, i
                    ):  # checks consistency with previously created objects
                        dist = math.sqrt(
                            (X0[l] - x0) ** 2 + (Y0[l] - y0) ** 2 + (Z0[l] - z0) ** 2
                        )
                        if (dist < (ab + AB[l])) or (
                            (abs(x0) + ab) ** 2
                            + (abs(y0) + ab) ** 2
                            + (abs(z0) + ab) ** 2
                            > 1.0
                        ):
                            breakj = True
                            break
                    if breakj == False:  # re-initialise if doesn't fit the criteria
                        X0[i] = x0
                        Y0[i] = y0
                        Z0[i] = z0
                        AB[i] = ab
                        C0_var[i] = c0
                        break
        if AB[i] == 0.0:
            X0[i] = x0
            Y0[i] = y0
            AB[i] = 0.0001
            C0_var[i] = c0

    myObjects = []  # dictionary of objects
    for obj in range(0, len(X0)):
        if mix_objects == True:
            rand_obj = random.randint(0, 2)
            if rand_obj == 0:
                object_type = Objects3D.ELLIPSOID
            if rand_obj == 1:
                object_type = Objects3D.PARABOLOID
            if rand_obj == 2:
                object_type = Objects3D.GAUSSIAN
        curr_obj = {
            "Obj": object_type,
            "C0": C0_var[obj],
            "x0": X0[obj],
            "y0": Y0[obj],
            "z0": Z0[obj],
            "a": AB[obj],
            "b": AB[obj],
            "c": AB[obj],
            "phi1": 0.0,
        }
        myObjects.append(curr_obj)
    Object3D = TomoP3D.Object(N_size, myObjects)
    return (Object3D, myObjects)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import ctypes
import ctypes.util
import sys, platform
import warnings


def c_shared_lib(lib_name, error=True):
    """Imports the library as a pre-built shared object"""
    if platform.system() == "Windows":
        load_dll = ctypes.windll.LoadLibrary
    else:
        load_dll = ctypes.cdll.LoadLibrary

    # Find the library and load it
    mylib_path = ctypes.util.find_library(lib_name)

    if mylib_path is not None:
        try:
            # No error if mylib_path is None; error if library name wrong
            return load_dll(mylib_path)
        except OSError:
            pass

    explanation = (
        "TomoPhantom links to compiled components which are installed separately"
        " and loaded using ctypes.util.find_library()."
    )
    if error:
        raise ModuleNotFoundError(
            explanation + f" A required library, {lib_name}, was not found."
        )
    warnings.warn(
        explanation + "Some functionality is unavailable because an optional shared"
        f" library, {lib_name}, is missing.",
        ImportWarning,
    )
    return None


from tomophantom.TomoP2D import *



import numpy as np
from numpy.testing import assert_allclose

import os
import tomophantom
from tomophantom.TomoP2D import Objects2D, Object
from tomophantom.supp.libraryToDict import modelfile2Dtolist

eps = 1e-05


def test_2d_object():
    N_size = 64  # set the desired dimension of the phantom
    # define all objects bellow:
    pp = {
        "Obj": Objects2D.GAUSSIAN,
        "C0": 1.00,
        "x0": 0.25,
        "y0": -0.3,
        "a": 0.15,
        "b": 0.3,
        "phi": -30.0,
    }

    pp1 = {
        "Obj": Objects2D.RECTANGLE,
        "C0": 1.00,
        "x0": -0.2,
        "y0": 0.2,
        "a": 0.25,
        "b": 0.4,
        "phi": 60.0,
    }

    myObjects = [pp, pp1]  # dictionary of objects
    Object1 = Object(N_size, myObjects)

    assert 0.0 <= np.max(Object1) <= 3
    assert Object1.dtype == np.float32
    assert Object1.shape == (N_size, N_size)


def test_2d_object_extraction():
    model = 11  # select a model number from the library
    N_size = 64  # set dimension of the phantom
    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library2D = os.path.join(path, "phantomlib", "Phantom2DLibrary.dat")
    # extract parameters into a list
    objlist = modelfile2Dtolist(path_library2D, model)
    # This will generate a N_size x N_size phantom (2D)
    phantom_2D = Object(N_size, objlist)

    assert 0.0 <= np.max(phantom_2D) <= 3
    assert phantom_2D.dtype == np.float32
    assert phantom_2D.shape == (N_size, N_size)



# this is a collection of speckle generating routines taken from:
# https://github.com/fperakis/Speckles-simulations

import numpy as np
from scipy.ndimage import gaussian_filter


def sample2ddist(asicshape, kbar, dist):
    """
    samples photons from a 2D distribution
    """
    # Definitions
    nasic = asicshape[0] * asicshape[1]
    nphotons = np.round(kbar * nasic).astype(int)
    photons = np.zeros(asicshape)

    # sample photon positions from given distribution
    sampl = np.random.choice(nasic, nphotons, p=dist)

    # convert to 2d and accumulate photons
    for i in range(nphotons):
        ix = int(sampl[i] / asicshape[1])
        iy = int(sampl[i] % asicshape[1])
        photons[ix, iy] += 1

    return photons


def model_speckles(asicshape, specklesize):
    """
    models a speckle pattern with a given speckle size (integer) based on random
    phasors (see JC Goodman - Speckle Phenomena in Optics - Appendix G)
    """
    # definitions speckles
    randphasors = np.zeros([asicshape[0], asicshape[1]], dtype=complex)
    nphasorsx = np.round(asicshape[0] / specklesize).astype(int)
    nphasorsy = np.round(asicshape[1] / specklesize).astype(int)

    # random phasors
    for i in range(nphasorsx):
        for j in range(nphasorsy):
            randphase = np.random.random_sample() * 2.0j * np.pi
            randphasors[i, j] = np.exp(randphase)

    # speckles
    specklefield = np.fft.fft2(randphasors)
    speckleint = np.absolute(specklefield) * np.absolute(specklefield)
    speckleint /= np.sum(speckleint)  # normalise

    return speckleint


def model_speckles_modes(asicshape, specklesize, modes):
    """
    models a speckle pattern with a given speckle size (integer) and
    contrast (1/modes)
    """

    # definitions
    nx, ny = asicshape[0], asicshape[1]

    # add speckle patterns as many times as the number of modes
    dist = model_speckles([nx, ny], specklesize)
    for m in range(modes - 1):
        dist += model_speckles([nx, ny], specklesize)
    dist /= np.sum(dist)  # normalise

    return dist


def simulate_speckles_with_shot_noise(asicshape, modes, specklesize, kbar):
    """
    returns a 2D speckle pattern with given speckle size, contrast
    (1/modes) and photon density kbar (photons/pixel).
    """
    # definitions
    nx, ny = asicshape[0], asicshape[1]
    dist = model_speckles_modes(asicshape, specklesize, modes).flatten()
    # sample photons from the distribution
    sig = sample2ddist([nx, ny], kbar, dist)

    return sig


def simulate_shot_noise(asicshape, kbar):
    """
    returns a 2d shot noise pattern with given dimensions and
    photon density kbar (photons/pixel)
    """
    # definitions
    nx, ny = asicshape[0], asicshape[1]

    # flat 2d distribution
    dist = np.ones((nx, ny)).flatten()
    dist /= np.sum(dist)

    # sample photons from the distribution
    sig = sample2ddist([nx, ny], kbar, dist)

    return np.array(sig, dtype=float)


def simulate_charge_sharing(asicshape, modes, specklesize, kbar, sigma=1):
    """
    returns a 2D speckle pattern with given speckle size, contrast
    (1/modes) and photon density kbar (photons/pixel)
    including charge sharing simulated with a gaussian at each photon position with with sigma
    """
    # Definitions
    nasic = asicshape[0] * asicshape[1]
    nphotons = np.round(kbar * nasic).astype(int)
    photons = np.zeros(asicshape)
    photons_blur = np.zeros(asicshape)
    dist = model_speckles_modes(asicshape, specklesize, modes).flatten()

    # sample photon positions from given distribution
    sampl = np.random.choice(nasic, nphotons, p=dist)

    # convert to 2d and accumulate photons
    for i in range(nphotons):
        ix = int(sampl[i] / asicshape[1])
        iy = int(sampl[i] % asicshape[1])
        photons_blur = np.zeros(asicshape)
        photons_blur[ix, iy] = 1
        photons_blur = gaussian_filter(photons_blur, sigma=sigma)
        photons += photons_blur

    return photons



# Defines common fixtures and makes them available to all tests

import os
import numpy as np
import pytest
import tomophantom


@pytest.fixture(scope="session")
def sino_model11_64():
    from tomophantom.TomoP2D import ModelSino

    N_size = 64  # set the desired dimension of the phantom
    model = 11
    angles_num = int(0.5 * np.pi * N_size)
    # angles number
    angles = np.linspace(0.0, 179.9, angles_num, dtype="float32")
    P = int(np.sqrt(2) * N_size)  # detectors

    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library2D = os.path.join(path, "phantomlib", "Phantom2DLibrary.dat")

    # Generate sinogram
    sino_an = ModelSino(model, N_size, P, angles, path_library2D)

    return (sino_an, angles_num, P)


@pytest.fixture(scope="session")
def sino_model11_3d_64():
    from tomophantom.TomoP3D import ModelSino

    N_size = 64  # set the desired dimension of the phantom
    # Projection geometry related parameters:
    Horiz_det = int(np.sqrt(2) * N_size)  # detector column count (horizontal)
    Vert_det = N_size  # detector row count (vertical) (no reason for it to be > N)
    angles_num = int(0.5 * np.pi * N_size)
    # angles number
    angles = np.linspace(0.0, 179.9, angles_num, dtype="float32")  # in degrees

    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library3D = os.path.join(path, "phantomlib", "Phantom3DLibrary.dat")

    # Generatea 3d sino
    sino3d = ModelSino(11, N_size, Horiz_det, Vert_det, angles, path_library3D)

    return (sino3d, Vert_det, angles_num, Horiz_det)



import numpy as np
from numpy.testing import assert_allclose

from tomophantom.TomoP3D import _check_params3d, Objects3D, Object

eps = 1e-05


def test_3d_object():
    N_size = 64
    obj3D_1 = {
        "Obj": Objects3D.GAUSSIAN,
        "C0": 1.0,
        "x0": -0.25,
        "y0": -0.15,
        "z0": 0.0,
        "a": 0.3,
        "b": 0.2,
        "c": 0.3,
        "phi1": 35.0,
    }

    obj3D_2 = {
        "Obj": Objects3D.CUBOID,
        "C0": 1.00,
        "x0": 0.1,
        "y0": 0.2,
        "z0": 0.0,
        "a": 0.15,
        "b": 0.35,
        "c": 0.6,
        "phi1": -60.0,
    }

    myObjects = [obj3D_1, obj3D_2]  # a list of dictionaries
    Object3D = Object(N_size, myObjects)

    assert 0.001 <= np.max(Object3D) <= 100
    assert Object3D.dtype == np.float32
    assert Object3D.shape == (N_size, N_size, N_size)



import numpy as np
from numpy.testing import assert_allclose

import pytest
import os
import tomophantom
from tomophantom.TomoP2D import _check_params2d, Model, ModelTemporal, SinoNum

eps = 1e-05


@pytest.mark.parametrize("model", list(np.arange(1, 16, 1)))
def test_2d_params_stat(model):
    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library2D = os.path.join(path, "phantomlib", "Phantom2DLibrary.dat")

    # check parameters
    params = _check_params2d(model, path_library2D)

    assert np.sum(params) == 10
    assert params.dtype == np.int32


@pytest.mark.parametrize("model", list(np.arange(100, 103, 1)))
def test_2d_params_temporal(model):
    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library2D = os.path.join(path, "phantomlib", "Phantom2DLibrary.dat")

    # check parameters
    params = _check_params2d(model, path_library2D)
    assert len(params) == 10
    if model == 100:
        assert params[3] == 3
    if model == 101:
        assert params[3] == 350
    if model == 102:
        assert params[3] == 25
    assert params.dtype == np.int32


@pytest.mark.parametrize("model", list(np.arange(1, 16, 1)))
def test_2d_models(model):
    N_size = 64  # set the desired dimension of the phantom
    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library2D = os.path.join(path, "phantomlib", "Phantom2DLibrary.dat")

    # Generate a N_size x N_size phantom (2D)
    phantom_2D = Model(model, N_size, path_library2D)
    assert 0.001 <= np.max(phantom_2D) <= 100
    assert phantom_2D.dtype == np.float32
    assert phantom_2D.shape == (N_size, N_size)


@pytest.mark.parametrize("model", list(np.arange(100, 103, 1)))
def test_2d_temporal_models(model):
    N_size = 64  # set the desired dimension of the phantom
    # one can specify an exact path to the parameters file
    path = os.path.dirname(tomophantom.__file__)
    path_library2D = os.path.join(path, "phantomlib", "Phantom2DLibrary.dat")

    params = _check_params2d(model, path_library2D)
    # Generate a time_frame x N_size x N_size phantom (3D)
    phantom_2D = ModelTemporal(model, N_size, path_library2D)
    assert 0.001 <= np.max(phantom_2D) <= 1000
    assert phantom_2D.dtype == np.float32
    assert phantom_2D.shape == (params[3], N_size, N_size)



import numpy as np

from tomophantom.flatsgen import synth_flats


def test_3d_models_sino(sino_model11_3d_64):
    sino3d, Vert_det, angles_num, Horiz_det = sino_model11_3d_64

    I0 = 50000
    # full-beam photon flux intensity
    flatsnum = 20  # the number of the flat fields required

    [projData3D_raw, flats_combined3D, speckel_map] = synth_flats(
        sino3d,
        source_intensity=I0,
        detectors_miscallibration=0.05,
        variations_number=3,
        arguments_Bessel=(1, 25),
        specklesize=2,
        kbar=2,
        jitter_projections=0.0,
        sigmasmooth=3,
        flatsnum=flatsnum,
    )

    assert 0.0 <= np.max(projData3D_raw) <= 65535
    assert projData3D_raw.dtype == np.uint16
    assert projData3D_raw.shape == (Vert_det, angles_num, Horiz_det)
    assert flats_combined3D.dtype == np.uint16
    assert flats_combined3D.shape == (Vert_det, flatsnum, Horiz_det)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Ctypes-wrapped C-modules of TomoPhantom library.
"""

import tomophantom.ctypes.dtype as dtype
from . import c_shared_lib

__all__ = [
    "c_checkParams2D",
    "c_model2d",
    "c_model_sino2d",
    "c_object_sino2d",
    "c_object2d",
    "c_checkParams3D",
    "c_model3d",
    "c_object3d",
    "c_model_sino3d",
    "c_object_sino3d",
]

LIB_TOMOPHANTOM = c_shared_lib("tomophantom")


def c_checkParams2D(
    params,
    model_no,
    models_library_path,
):
    LIB_TOMOPHANTOM.checkParams2D.restype = dtype.as_c_void_p()
    LIB_TOMOPHANTOM.checkParams2D(
        dtype.as_c_int_p(params),
        dtype.as_c_int(model_no),
        dtype.as_c_char_p(models_library_path),
    )
    return params


def c_checkParams3D(
    params,
    model_no,
    models_library_path,
):
    LIB_TOMOPHANTOM.checkParams3D.restype = dtype.as_c_void_p()
    LIB_TOMOPHANTOM.checkParams3D(
        dtype.as_c_int_p(params),
        dtype.as_c_int(model_no),
        dtype.as_c_char_p(models_library_path),
    )
    return params


def c_numerical_sino2d(
    sinogram,
    input,
    phantom_size,
    detector_X,
    angles,
    angles_total,
):
    LIB_TOMOPHANTOM.TomoP2DSinoNum_core.restype = dtype.as_c_void_p()
    LIB_TOMOPHANTOM.TomoP2DSinoNum_core(
        dtype.as_c_float_p(sinogram),
        dtype.as_c_float_p(input),
        dtype.as_c_int(phantom_size),
        dtype.as_c_int(detector_X),
        dtype.as_c_float_p(angles),
        dtype.as_c_int(angles_total),
        0,
    )
    return sinogram


def c_model2d(
    output,
    model_no,
    phantom_size,
    models_library_path,
):
    LIB_TOMOPHANTOM.TomoP2DModel_core.restype = dtype.as_c_void_p()
    LIB_TOMOPHANTOM.TomoP2DModel_core(
        dtype.as_c_float_p(output),
        dtype.as_c_int(model_no),
        dtype.as_c_int(phantom_size),
        dtype.as_c_char_p(models_library_path),
    )
    return output


def c_model_sino2d(
    sinogram,
    model_no,
    phantom_size,
    detector_size,
    angles,
    angles_total,
    models_library_path,
):
    LIB_TOMOPHANTOM.TomoP2DModelSino_core.restype = dtype.as_c_void_p()
    LIB_TOMOPHANTOM.TomoP2DModelSino_core(
        dtype.as_c_float_p(sinogram),
        dtype.as_c_int(model_no),
        dtype.as_c_int(phantom_size),
        dtype.as_c_int(detector_size),
        dtype.as_c_float_p(angles),
        dtype.as_c_int(angles_total),
        1,
        dtype.as_c_char_p(models_library_path),
    )
    return sinogram


def c_object_sino2d(
    sinogram,
    phantom_size,
    detector_size,
    angles,
    angles_total,
    objectName,
    C0,
    y0,
    x0,
    a,
    b,
    phi_ang,
    tt,
):
    LIB_TOMOPHANTOM.TomoP2DObjectSino_core.restype = dtype.as_c_void_p()
    LIB_TOMOPHANTOM.TomoP2DObjectSino_core(
        dtype.as_c_float_p(sinogram),
        dtype.as_c_int(phantom_size),
        dtype.as_c_int(detector_size),
        dtype.as_c_float_p(angles),
        dtype.as_c_int(angles_total),
        1,
        dtype.as_c_char_p(objectName),
        dtype.as_c_float(C0),
        dtype.as_c_float(y0),
        dtype.as_c_float(x0),
        dtype.as_c_float(a),
        dtype.as_c_float(b),
        dtype.as_c_float(phi_ang),
        dtype.as_c_int(tt),
    )
    return sinogram


def c_object2d(object2d, phantom_size, objectName, C0, y0, x0, a, b, phi_ang, tt):
    LIB_TOMOPHANTOM.TomoP2DObject_core.restype = dtype.as_c_void_p()
    LIB_TOMOPHANTOM.TomoP2DObject_core(
        dtype.as_c_float_p(object2d),
        dtype.as_c_int(phantom_size),
        dtype.as_c_char_p(objectName),
        dtype.as_c_float(C0),
        dtype.as_c_float(x0),
        dtype.as_c_float(y0),
        dtype.as_c_float(b),
        dtype.as_c_float(a),
        dtype.as_c_float(phi_ang),
        dtype.as_c_int(tt),
    )
    return object2d


def c_model3d(
    output,
    model_no,
    N1,
    N2,
    N3,
    Z1,
    Z2,
    models_library_path,
):
    LIB_TOMOPHANTOM.TomoP3DModel_core.restype = dtype.as_c_void_p()
    LIB_TOMOPHANTOM.TomoP3DModel_core(
        dtype.as_c_float_p(output),
        dtype.as_c_int(model_no),
        dtype.as_c_long(N3),
        dtype.as_c_long(N2),
        dtype.as_c_long(N1),
        dtype.as_c_long(Z1),
        dtype.as_c_long(Z2),
        dtype.as_c_char_p(models_library_path),
    )
    return output


def c_object3d(
    output,
    N1,
    N2,
    N3,
    Z1,
    Z2,
    objectName,
    C0,
    x0,
    y0,
    z0,
    a,
    b,
    c,
    phi1,
    phi2,
    phi3,
    tt,
):
    LIB_TOMOPHANTOM.TomoP3DObject_core.restype = dtype.as_c_void_p()
    LIB_TOMOPHANTOM.TomoP3DObject_core(
        dtype.as_c_float_p(output),
        dtype.as_c_long(N3),
        dtype.as_c_long(N2),
        dtype.as_c_long(N1),
        dtype.as_c_long(Z1),
        dtype.as_c_long(Z2),
        dtype.as_c_char_p(objectName),
        dtype.as_c_float(C0),
        dtype.as_c_float(y0),
        dtype.as_c_float(x0),
        dtype.as_c_float(z0),
        dtype.as_c_float(a),
        dtype.as_c_float(b),
        dtype.as_c_float(c),
        dtype.as_c_float(phi1),
        dtype.as_c_float(phi2),
        dtype.as_c_float(phi3),
        dtype.as_c_long(tt),
    )
    return output


def c_model_sino3d(
    output,
    model_no,
    detector_horiz,
    detector_vert,
    Z1,
    Z2,
    phantom_size,
    angles,
    angles_total,
    models_library_path,
):
    LIB_TOMOPHANTOM.TomoP3DModelSino_core.restype = dtype.as_c_void_p()
    LIB_TOMOPHANTOM.TomoP3DModelSino_core(
        dtype.as_c_float_p(output),
        dtype.as_c_int(model_no),
        dtype.as_c_long(detector_horiz),
        dtype.as_c_long(detector_vert),
        dtype.as_c_long(Z1),
        dtype.as_c_long(Z2),
        dtype.as_c_long(phantom_size),
        dtype.as_c_float_p(angles),
        dtype.as_c_int(angles_total),
        dtype.as_c_char_p(models_library_path),
    )
    return output


def c_object_sino3d(
    output,
    detector_horiz,
    detector_vert,
    Z1,
    Z2,
    phantom_size,
    angles,
    angles_total,
    objectName,
    C0,
    x0,
    y0,
    z0,
    a,
    b,
    c,
    phi1,
    phi2,
    phi3,
    tt,
):
    LIB_TOMOPHANTOM.TomoP3DObjectSino_core.restype = dtype.as_c_void_p()
    if (
        ("gaussian" in str(objectName))
        or ("paraboloid" in str(objectName))
        or ("ellipsoid" in str(objectName))
    ):
        LIB_TOMOPHANTOM.TomoP3DObjectSino_core(
            dtype.as_c_float_p(output),
            dtype.as_c_long(detector_horiz),
            dtype.as_c_long(detector_vert),
            dtype.as_c_long(Z1),
            dtype.as_c_long(Z2),
            dtype.as_c_long(phantom_size),
            dtype.as_c_float_p(angles),
            dtype.as_c_int(angles_total),
            dtype.as_c_char_p(objectName),
            dtype.as_c_float(C0),
            dtype.as_c_float(y0),
            dtype.as_c_float(-z0),
            dtype.as_c_float(-x0),
            dtype.as_c_float(b),
            dtype.as_c_float(a),
            dtype.as_c_float(c),
            dtype.as_c_float(phi3),
            dtype.as_c_float(phi2),
            dtype.as_c_float(phi1),
            dtype.as_c_long(tt),
        )
    elif "elliptical_cylinder" in str(objectName):
        LIB_TOMOPHANTOM.TomoP3DObjectSino_core(
            dtype.as_c_float_p(output),
            dtype.as_c_long(detector_horiz),
            dtype.as_c_long(detector_vert),
            dtype.as_c_long(Z1),
            dtype.as_c_long(Z2),
            dtype.as_c_long(phantom_size),
            dtype.as_c_float_p(angles),
            dtype.as_c_int(angles_total),
            dtype.as_c_char_p(objectName),
            dtype.as_c_float(C0),
            dtype.as_c_float(x0),
            dtype.as_c_float(-y0),
            dtype.as_c_float(z0),
            dtype.as_c_float(b),
            dtype.as_c_float(a),
            dtype.as_c_float(c),
            dtype.as_c_float(phi3),
            dtype.as_c_float(phi2),
            dtype.as_c_float(phi1),
            dtype.as_c_long(tt),
        )
    else:
        LIB_TOMOPHANTOM.TomoP3DObjectSino_core(
            dtype.as_c_float_p(output),
            dtype.as_c_long(detector_horiz),
            dtype.as_c_long(detector_vert),
            dtype.as_c_long(Z1),
            dtype.as_c_long(Z2),
            dtype.as_c_long(phantom_size),
            dtype.as_c_float_p(angles),
            dtype.as_c_int(angles_total),
            dtype.as_c_char_p(objectName),
            dtype.as_c_float(C0),
            dtype.as_c_float(x0),
            dtype.as_c_float(y0),
            dtype.as_c_float(z0),
            dtype.as_c_float(a),
            dtype.as_c_float(b),
            dtype.as_c_float(c),
            dtype.as_c_float(phi3),
            dtype.as_c_float(phi2),
            dtype.as_c_float(-phi1),
            dtype.as_c_long(tt),
        )
    return output



import numpy as np
from numpy.testing import assert_allclose

from tomophantom.TomoP3D import Objects3D, ObjectSino


def test_3d_object_sino():
    N3D_size = 64  # set the desired dimension of the phantom
    Horiz_det = int(np.sqrt(2) * N3D_size)  # detector column count (horizontal)
    Vert_det = N3D_size  # detector row count (vertical) (no reason for it to be > N)
    angles_num = int(0.5 * np.pi * N3D_size)
    # angles number
    angles = np.linspace(0.0, 179.9, angles_num, dtype="float32")  # in degrees

    # specify object parameters, here we replicate model
    obj3D_1 = {
        "Obj": Objects3D.GAUSSIAN,
        "C0": 1.0,
        "x0": -0.25,
        "y0": -0.15,
        "z0": 0.0,
        "a": 0.3,
        "b": 0.2,
        "c": 0.3,
        "phi1": 35.0,
    }

    obj3D_2 = {
        "Obj": Objects3D.CUBOID,
        "C0": 1.00,
        "x0": 0.1,
        "y0": 0.2,
        "z0": 0.0,
        "a": 0.15,
        "b": 0.35,
        "c": 0.6,
        "phi1": -60.0,
    }

    myObjects = [obj3D_1, obj3D_2]  # dictionary of objects
    ProjData3D = ObjectSino(N3D_size, Horiz_det, Vert_det, angles, myObjects)

    assert 0.0 <= np.max(ProjData3D) <= 1000
    assert ProjData3D.dtype == np.float32
    assert ProjData3D.shape == (Vert_det, angles_num, Horiz_det)



# -*- coding: utf-8 -*-
"""
Created on Thu Mar 15 11:12:51 2018

@author: ofn77899
"""
import re
from tomophantom.TomoP2D import Objects2D


def modelfile2Dtolist(filepath, model):
    # read all the file content
    f = open(filepath).readlines()

    # model 4 is lines 56
    i = 0
    model = model
    components = 0
    #    timesteps = 1
    while True:
        if f[i] == "Model : {0:02};\n".format(model):
            components = int(re.search("^Components : (\d+);$", f[i + 1]).groups()[0])
            #            timesteps = int( re.search('^TimeSteps : (\d+);$',f[i+2]).groups()[0] )
            #            print (i)
            i += 3
            #            print (i)
            #            print (f[i])
            break
        i += 1
    # print (i)

    objectlist = []

    for j in range(components):
        descr = f[i + j].split()
        objstr, C0, x0, y0, a, b, phi = descr[2:]
        for oo in Objects2D:
            if oo.value == objstr:
                break
        #        print (phi)
        objectlist.append(
            {
                "Obj": oo,
                "C0": float(C0),
                "x0": float(y0),
                "y0": float(x0),
                "a": float(b),
                "b": float(a),
                "phi": float(phi[:-1]),
            }
        )
    return objectlist


if __name__ == "__main__":
    pathTP = "../phantomlib/Phantom2DLibrary.dat"

    print(modelfile2Dtolist(pathTP, 6))
